import datetime
from typing import Any, List, Dict, Union, Optional
import numpy as np  # type: ignore
import xarray as xr  # type: ignore

Timetype = Union[str, np.datetime64, datetime.datetime]
Array = Union[np.ndarray, xr.DataArray]


class InstanceVariable:
    def __init__(
        self,
        data: xr.DataArray,
        pid: xr.DataArray,
        ptime: xr.DataArray,
        pcount: np.ndarray,
    ) -> None:
        self.da = data
        self.pid = pid
        self.time = ptime
        self.count = pcount
        self.end = self.count.cumsum()
        self.start = self.end - self.count
        self.num_times = len(self.time)
        self.particles = np.unique(self.pid)
        self.num_particles = len(self.particles)  # Number of distinct particles

    # @property
    # def end(self) -> np.ndarray:
    #    return self.count.cumsum()

    # @property
    # def start(self) -> np.ndarray:
    #    return self.end - self.count

    @property
    def values(self) -> np.ndarray:
        # Same as self.da.values
        return np.array(self.da)

    def _sel_time_index(self, n: int) -> xr.DataArray:
        """Select by time index, return xarray."""
        start = self.start[n]
        end = self.end[n]
        V = self.da[start:end]
        V = V.assign_coords(time=self.time[n])
        V = V.assign_coords(pid=self.pid[start:end])
        V = V.swap_dims({"particle_instance": "pid"})
        return V

    def _sel_time_slice_index(self, tslice: slice) -> "InstanceVariable":
        """Take a time slice based on time indices"""
        n = self.num_times
        istart, istop, step = tslice.indices(n)
        if step != 1:
            raise IndexError("step > 1 is not allowed")
        start = self.start[istart]
        end = self.end[istop - 1]
        return InstanceVariable(
            data=self.da[start:end],
            pid=self.pid[start:end],
            ptime=self.time[tslice],
            pcount=self.count[tslice],
        )

    def _sel_time_value(self, time_val: Timetype) -> xr.DataArray:
        idx = self.time.get_index("time").get_loc(time_val)
        return self._sel_time_index(idx)

    def _sel_pid_value(self, pid: int) -> xr.DataArray:
        """Selection based on single pid value"""
        # Burde få en pid-koordinat også
        data = []
        times = []
        for t_idx in range(self.num_times):
            try:
                data.append(self._sel_time_index(t_idx).sel(pid=pid))
                times.append(t_idx)
            except KeyError:
                pass
        # Bedre, på forhånd sjekk om pid > maximum
        if not data:
            raise KeyError(f"No such pid = {pid}")
        V = xr.DataArray(data, coords={"time": self.time[times]}, dims=("time",))
        V["pid"] = pid
        return V

    # def isel(self, *, time: Optional[int] = None) -> xr.DataArray:
    #     if time is not None:
    #         return self._sel_time_index(time)
    #     else:
    #         raise ValueError("Need one argument")
    def isel(self, *, time: int) -> xr.DataArray:
        return self._sel_time_index(time)

    def sel(
        self, *, pid: Optional[int] = None, time: Optional[Timetype] = None
    ) -> xr.DataArray:
        """Select from InstanceVariable by value of pid or time or both"""
        if pid is not None and time is None:
            return self._sel_pid_value(pid)
        if time is not None and pid is None:
            return self._sel_time_value(time)
        if time is not None and pid is not None:
            return self._sel_time_value(time).sel(pid=pid)
        # No arguments
        raise ValueError("Need 1 or 2 arguments")

    # Do something like dask if the array gets to big
    def full(self) -> xr.DataArray:
        """Return a full DataArray"""
        data = np.empty((self.num_times, self.num_particles))
        data[:, :] = np.nan
        for n in range(self.num_times):
            data[n, self.pid[self.start[n] : self.end[n]]] = self._sel_time_index(n)
        # coords = dict(time=self.time, pid=self.particles)
        coords = [("time", self.time), ("pid", self.particles)]
        V = xr.DataArray(data=data, coords=coords, dims=("time", "pid"))
        return V

    # More complicated typing
    # def __getitem__(self, index: Union[int, slice]) -> xr.DataArray:
    def __getitem__(
        self, index: Union[int, slice]
    ) -> Union[xr.DataArray, "InstanceVariable"]:
        if isinstance(index, int):  # index = time_idx
            return self._sel_time_index(index)
        if isinstance(index, slice):
            return self._sel_time_slice_index(index)
        else:  # index = time_idx, pid
            time_idx, pid = index
            if 0 <= pid < self.num_particles:
                try:
                    v = self._sel_time_index(time_idx).sel(pid=pid)
                except KeyError:
                    # Også håndtere v != floatpf.
                    v = np.nan
            else:
                raise IndexError(f"pid={pid} is out of bound={self.num_particles}")
            return v

    def __array__(self) -> np.ndarray:
        return np.array(self.da)

    def __repr__(self) -> str:
        s = "<postladim.InstanceVariable>\n"
        s += f"num_times: {self.num_times}, particle_instance: {len(self.da)}\n"
        s += arraystr(self.da)
        return s

    def __len__(self) -> int:
        return len(self.time)


# --------------------------------------------

# Need this?, just use the DataArray
class ParticleVariable:
    """Particle variable, time-independent"""

    # def __init__(self, particlefile: "ParticleFile", varname: str) -> None:
    def __init__(self, data: xr.DataArray) -> None:
        self.da = data

    def __getitem__(self, p: int) -> Any:
        """Get the value of particle with pid = p
        """
        return self.da[p]

    def __array__(self) -> np.ndarray:
        return np.array(self.da)

    def __repr__(self) -> str:
        s = "<postladim.ParticleVariable>\n"
        s += f"particle: {len(self.da)}\n"
        s += arraystr(self.da)
        return s

    def __len__(self) -> int:
        return len(self.da)


def itemstr(v: Array) -> str:
    """Pretty print array item"""

    # Date
    if str(v.dtype).startswith("datetime64"):
        return str(v.__array__()).rstrip("0.:T")

    # Number
    return f"{v:g}"


def arraystr(A: Array) -> str:
    """Pretty print array"""
    B = np.asarray(A).ravel()
    if len(B) <= 3:
        return " ".join([itemstr(v) for v in B])
    return " ".join([itemstr(B[0]), itemstr(B[1]), "...", itemstr(B[-1])])



"""
Grid and Forcing for LADiM for the Regional Ocean Model System (ROMS)

"""

# -----------------------------------
# Bjørn Ådlandsvik, <bjorn@imr.no>
# Institute of Marine Research
# Bergen, Norway
# 2017-03-01
# -----------------------------------

# import sys
import glob
import logging
import numpy as np
from netCDF4 import Dataset, num2date
from ladim.sample import sample2D


class Grid:
    """Simple ROMS grid object

    Possible grid arguments:
      subgrid = [i0, i1, j0, j1]
        Ordinary python style, start points included, not end points
        Each of the elements can be replaced with None, for no limitation
      Vinfo: dictionary with N, hc, theta_s and theta_b

    """

    # Lagrer en del unødige attributter

    def __init__(self, config):

        logging.info("Initializing zROMS grid object")
        try:
            ncid = Dataset(config["gridforce"]["grid_file"])
        except OSError:
            logging.error(
                "Grid file {} not found".format(config["gridforce"]["grid_file"])
            )
            raise SystemExit(1)

        # Subgrid, only considers internal grid cells
        # 1 <= i0 < i1 <= imax-1, default=end points
        # 1 <= j0 < j1 <= jmax-1, default=end points
        # Here, imax, jmax refers to whole grid
        jmax, imax = ncid.variables["h"].shape
        whole_grid = [1, imax - 1, 1, jmax - 1]
        if "subgrid" in config["gridforce"]:
            limits = list(config["gridforce"]["subgrid"])
        else:
            limits = whole_grid
        # Allow None if no imposed limitation
        for ind, val in enumerate(limits):
            if val is None:
                limits[ind] = whole_grid[ind]

        self.i0, self.i1, self.j0, self.j1 = limits
        self.imax = self.i1 - self.i0
        self.jmax = self.j1 - self.j0
        # print('Grid : imax, jmax, size = ',
        #      self.imax, self.jmax, self.imax*self.jmax)

        # Limits for where velocities are defined
        self.xmin = float(self.i0)
        self.xmax = float(self.i1 - 1)
        self.ymin = float(self.j0)
        self.ymax = float(self.j1 - 1)

        # Slices
        #   rho-points
        self.I = slice(self.i0, self.i1)
        self.J = slice(self.j0, self.j1)
        #   U and V-points
        self.Iu = slice(self.i0 - 1, self.i1)
        self.Ju = self.J
        self.Iv = self.I
        self.Jv = slice(self.j0 - 1, self.j1)

        # Vertical z-levels
        # self.z_levels = ncid.variables['zlevels'][:]
        self.z_levels = np.array([0, 1, 2, 3, 5, 10, 15, 20])

        # Read some variables
        self.H = ncid.variables["h"][self.J, self.I]
        self.M = ncid.variables["mask_rho"][self.J, self.I].astype(int)
        # self.Mu = ncid.variables['mask_u'][self.Ju, self.Iu]
        # self.Mv = ncid.variables['mask_v'][self.Jv, self.Iv]
        self.dx = 1.0 / ncid.variables["pm"][self.J, self.I]
        self.dy = 1.0 / ncid.variables["pn"][self.J, self.I]
        self.lon = ncid.variables["lon_rho"][self.J, self.I]
        self.lat = ncid.variables["lat_rho"][self.J, self.I]

        # self.z_r = sdepth(self.H, self.hc, self.Cs_r,
        #                  stagger='rho', Vtransform=self.Vtransform)
        # self.z_w = sdepth(self.H, self.hc, self.Cs_w,
        #                  stagger='w', Vtransform=self.Vtransform)

        # Land masks at u- and v-points
        M = self.M
        Mu = np.zeros((self.jmax, self.imax + 1), dtype=int)
        Mu[:, 1:-1] = M[:, :-1] * M[:, 1:]
        Mu[:, 0] = M[:, 0]
        Mu[:, -1] = M[:, -1]
        self.Mu = Mu

        Mv = np.zeros((self.jmax + 1, self.imax), dtype=int)
        Mv[1:-1, :] = M[:-1, :] * M[1:, :]
        Mv[0, :] = M[0, :]
        Mv[-1, :] = M[-1, :]
        self.Mv = Mv

        # Close the file(s)
        ncid.close()

    def sample_metric(self, X, Y):
        """Sample the metric coefficients

        Changes slowly, so using nearest neighbour
        """
        I = X.round().astype(int) - self.i0
        J = Y.round().astype(int) - self.j0

        # Metric is conform for PolarStereographic
        A = self.dx[J, I]
        return A, A

    def sample_depth(self, X, Y):
        """Return the depth of grid cells"""
        I = X.round().astype(int) - self.i0
        J = Y.round().astype(int) - self.j0
        return self.H[J, I]

    def lonlat(self, X, Y, method="bilinear"):
        """Return the longitude and latitude from grid coordinates"""
        if method == "bilinear":  # More accurate
            return self.xy2ll(X, Y)
        else:  # containing grid cell, less accurate
            I = X.round().astype("int") - self.i0
            J = Y.round().astype("int") - self.j0
            return self.lon[J, I], self.lat[J, I]

        return (
            sample2D(self.lon, X - self.i0, Y - self.j0),
            sample2D(self.lat, X - self.i0, Y - self.j0),
        )

    def ingrid(self, X, Y):
        """Returns True for points inside the subgrid"""
        return (self.xmin < X) & (X < self.xmax) & (self.ymin < Y) & (Y < self.ymax)

    def onland(self, X, Y):
        """Returns True for points on land"""
        I = X.round().astype(int) - self.i0
        J = Y.round().astype(int) - self.j0
        return self.M[J, I] < 1

    # Error if point outside
    def atsea(self, X, Y):
        """Returns True for points at sea"""
        I = X.round().astype(int) - self.i0
        J = Y.round().astype(int) - self.j0
        return self.M[J, I] > 0

    def xy2ll(self, X, Y):
        return (
            sample2D(self.lon, X - self.i0, Y - self.j0),
            sample2D(self.lat, X - self.i0, Y - self.j0),
        )

    def ll2xy(self, lon, lat):
        Y, X = bilin_inv(lon, lat, self.lon, self.lat)
        return X + self.i0, Y + self.j0


# -----------------------------------------------
# The Forcing class from the old forcing module
# -----------------------------------------------


class Forcing:
    """
    Class for ROMS forcing

    """

    def __init__(self, config, grid):

        logging.info("Initiating forcing")

        self._grid = grid  # Get the grid object, make private?

        self.ibm_forcing = config["ibm_forcing"]

        # Test for glob, use MFDataset if needed
        files = glob.glob(config["gridforce"]["input_file"])
        files.sort()
        numfiles = len(files)
        if numfiles == 0:
            logging.error("No input file: {}".format(config["gridforce"]["input_file"]))
            raise SystemExit(3)
        logging.info("Number of forcing files = {}".format(numfiles))

        # ----------------------------------------
        # Open first file for some general info
        # must be valid for all the files
        # ----------------------------------------

        with Dataset(files[0]) as nc:

            # time_units = nc.variables['ocean_time'].units
            time_units = nc.variables["time"].units

            self.scaled = dict()
            self.scale_factor = dict()
            self.add_offset = dict()

            if hasattr(nc.variables["u"], "scale_factor"):
                self.scaled["U"] = True
                self.scale_factor["U"] = np.float32(nc.variables["u"].scale_factor)
                self.add_offset["U"] = np.float32(nc.variables["u"].add_offset)
                self.scaled["V"] = True
                self.scale_factor["V"] = np.float32(self.scale_factor["U"])
                self.add_offset["V"] = np.float32(self.add_offset["U"])
            else:
                self.scaled["U"] = False
                self.scaled["V"] = False

            for key in self.ibm_forcing:
                if hasattr(nc.variables[key], "scale_factor"):
                    self.scaled[key] = True
                    self.scale_factor[key] = np.float32(nc.variables[key].scale_factor)
                    self.add_offset[key] = np.float32(nc.variables[key].add_offset)
                else:
                    self.scaled[key] = False

        # ---------------------------
        # Overview of all the files
        # ---------------------------

        times = []  # list of times of all frames
        num_frames = []  # Available time frames in each file
        # change_times = []     # Times for change of file
        for fname in files:
            print(fname)
            with Dataset(fname) as nc:
                # new_times = nc.variables['ocean_time'][:]
                new_times = nc.variables["time"][:]
                times.extend(new_times)
                num_frames.append(len(new_times))
        logging.info("Number of available forcing times = {:d}".format(len(times)))

        # Find first/last forcing times
        # -----------------------------
        time0 = num2date(times[0], time_units)
        time1 = num2date(times[-1], time_units)
        logging.info("time0 = {}".format(str(time0)))
        logging.info("time1 = {}".format(str(time1)))
        # print(time0)
        # print(time1)
        start_time = np.datetime64(config["start_time"])
        # self.time = start_time
        self.dt = np.timedelta64(int(config["dt"]), "s")  # or use

        # Check that forcing period covers the simulation period
        # ------------------------------------------------------
        # Use logging module for this

        if time0 > start_time:
            logging.error("No forcing at start time")
            raise SystemExit(3)
        if time1 < config["stop_time"]:
            logging.error("No forcing at stop time")
            raise SystemExit(3)

        # Make a list steps of the forcing time steps
        # --------------------------------------------
        steps = []  # Model time step of forcing
        for t in times:
            # print(num2date(t, time_units), t)
            otime = np.datetime64(str(num2date(t, time_units)))
            dtime = np.timedelta64(otime - start_time, "s").astype(int)
            steps.append(int(dtime / config["dt"]))

        # print(steps)

        # change_steps = [steps[t] for t in change_times]

        file_idx = dict()
        frame_idx = dict()
        step_counter = -1
        for i, fname in enumerate(files):
            for frame in range(num_frames[i]):
                step_counter += 1
                step = steps[step_counter]
                # print(step_counter, step, i, frame)
                file_idx[step] = i
                frame_idx[step] = frame

        self._files = files
        self.stepdiff = np.diff(steps)
        self.file_idx = file_idx
        self.frame_idx = frame_idx

        # Read old input
        # requires at least one input before start
        # to get Runge-Kutta going
        # --------------
        # prestep = last forcing step < 0
        #
        V = [step for step in steps if step < 0]
        if V:  # Forcing available before start time
            prestep = max(V)
            stepdiff = self.stepdiff[steps.index(prestep)]
            nextstep = prestep + stepdiff
            self.U, self.V = self._read_velocity(prestep)
            self.Unew, self.Vnew = self._read_velocity(nextstep)
            self.dU = (self.Unew - self.U) / stepdiff
            self.dV = (self.Vnew - self.V) / stepdiff
            # Interpolate to time step = -1
            self.U = self.U - (prestep + 1) * self.dU
            self.V = self.V - (prestep + 1) * self.dV
            # Other forcing
            for name in self.ibm_forcing:
                self[name] = self._read_field(name, prestep)
                self[name + "new"] = self._read_field(name, nextstep)
                self["d" + name] = (self[name + "new"] - self[name]) / prestep
                self[name] = self[name] - (prestep + 1) * self["d" + name]

        elif steps[0] == 0:
            # Simulation start at first forcing time
            self.U, self.V = self._read_velocity(0)
            self.Unew, self.Vnew = self._read_velocity(steps[1])
            self.dU = (self.Unew - self.U) / steps[1]
            self.dV = (self.Vnew - self.V) / steps[1]
            # Syncronize
            self.Unew = self.U
            self.Vnew = self.V
            # Extrapolate to time step = -1
            self.U = self.U - self.dU
            self.V = self.V - self.dV
            # Other forcing:
            for name in self.ibm_forcing:
                self[name] = self._read_field(name, 0)
                self[name + "new"] = self._read_field(name, steps[1])
                self["d" + name] = (self[name + "new"] - self[name]) / steps[1]
                self[name] = self[name] - self["d" + name]

        self.steps = steps
        self._files = files

    # ==============================================

    # Turned off time interpolation of scalar fields
    # TODO: Implement a switch for turning it on again if wanted
    def update(self, t):
        """Update the fields to time step t"""

        # Read from config
        interpolate_velocity_in_time = True
        interpolate_ibm_forcing_in_time = False

        logging.debug("Updating forcing, time step = {}".format(t))
        if t in self.steps:  # No time interpolation
            self.U = self.Unew
            self.V = self.Vnew
            for name in self.ibm_forcing:
                self[name] = self[name + "new"]
        else:
            if t - 1 in self.steps:  # Need new fields
                stepdiff = self.stepdiff[self.steps.index(t - 1)]
                nextstep = t - 1 + stepdiff
                self.Unew, self.Vnew = self._read_velocity(nextstep)
                for name in self.ibm_forcing:
                    self[name + "new"] = self._read_field(name, nextstep)
                if interpolate_velocity_in_time:
                    self.dU = (self.Unew - self.U) / stepdiff
                    self.dV = (self.Vnew - self.V) / stepdiff
                if interpolate_ibm_forcing_in_time:
                    for name in self.ibm_forcing:
                        self["d" + name] = (self[name + "new"] - self[name]) / stepdiff

            # "Ordinary" time step (including self.steps+1)
            if interpolate_velocity_in_time:
                self.U += self.dU
                self.V += self.dV
            if interpolate_ibm_forcing_in_time:
                for name in self.ibm_forcing:
                    self[name] += self["d" + name]

    # --------------

    def _read_velocity(self, n):
        """Read fields at time step = n"""
        # Need a switch for reading W
        # T = self._nc.variables['ocean_time'][n]  # Read new fields

        # Handle file opening/closing
        # Always read velocity before other fields
        logging.debug("Reading velocity for time step = {}".format(n))
        first = True
        if first:  # Open file initiallt
            self._nc = Dataset(self._files[self.file_idx[n]])
            self._nc.set_auto_maskandscale(False)
            first = False
        else:
            if self.frame_idx[n] == 0:  # New file
                self._nc.close()  # Close previous file
                self._nc = Dataset(self._files[self.file_idx[n]])
                self._nc.set_auto_maskandscale(False)

        frame = self.frame_idx[n]

        # Read the velocity
        U = self._nc.variables["u"][frame, :, self._grid.Ju, self._grid.Iu]
        V = self._nc.variables["v"][frame, :, self._grid.Jv, self._grid.Iv]
        # Scale if needed
        # Assume offset = 0 for velocity
        if self.scaled["U"]:
            U = self.scale_factor["U"] * U
            V = self.scale_factor["U"] * V
            # U = self.add_offset['U'] + self.scale_factor['U']*U
            # V = self.add_offset['U'] + self.scale_factor['U']*V

        # If necessary put U,V = zero on land and land boundaries
        # Stay as float32
        np.multiply(U, self._grid.Mu, out=U)
        np.multiply(V, self._grid.Mv, out=V)
        return U, V

    def _read_field(self, name, n):
        """Read a 3D field"""
        # print("IBM-forcing:", name)
        frame = self.frame_idx[n]
        F = self._nc.variables[name][frame, :, self._grid.J, self._grid.I]
        if self.scaled[name]:
            F = self.add_offset[name] + self.scale_factor[name] * F
        return F

    # Allow item notation
    def __setitem__(self, key, value):
        setattr(self, key, value)

    def __getitem__(self, key):
        return getattr(self, key)

    # ------------------

    def close(self):

        self._nc.close()

    # def sample_velocity(self, X, Y, Z, tstep=0, method='bilinear'):
    def velocity(self, X, Y, Z, tstep=0, method="bilinear"):

        i0 = self._grid.i0
        j0 = self._grid.j0
        K, A = vert_level(self._grid.z_levels, X - i0, Y - j0, Z)
        if tstep < 0.001:
            return sample3DUV(self.U, self.V, X - i0, Y - j0, K, A, method=method)
        else:
            return sample3DUV(
                self.U + tstep * self.dU,
                self.V + tstep * self.dV,
                X - i0,
                Y - j0,
                K,
                A,
                method=method,
            )

    # Simplify to grid cell
    def field(self, X, Y, Z, name):
        # should not be necessary to repeat
        i0 = self._grid.i0
        j0 = self._grid.j0
        K, A = vert_level(self._grid.z_levels, X - i0, Y - j0, Z)
        F = self[name]
        return sample3D(F, X - i0, Y - j0, K, A, method="nearest")


# ---------------------------------------------
#      Low-level vertical functions
#      more or less from the roppy package
#      https://github.com/bjornaa/roppy
# ----------------------------------------------


def s_stretch(N, theta_s, theta_b, stagger="rho", Vstretching=1):
    """Compute a s-level stretching array

    *N* : Number of vertical levels

    *theta_s* : Surface stretching factor

    *theta_b* : Bottom stretching factor

    *stagger* : "rho"|"w"

    *Vstretching* : 1|2|4

    """

    if stagger == "rho":
        S = -1.0 + (0.5 + np.arange(N)) / N
    elif stagger == "w":
        S = np.linspace(-1.0, 0.0, N + 1)
    else:
        raise ValueError("stagger must be 'rho' or 'w'")

    if Vstretching == 1:
        cff1 = 1.0 / np.sinh(theta_s)
        cff2 = 0.5 / np.tanh(0.5 * theta_s)
        return (1.0 - theta_b) * cff1 * np.sinh(theta_s * S) + theta_b * (
            cff2 * np.tanh(theta_s * (S + 0.5)) - 0.5
        )

    elif Vstretching == 2:
        a, b = 1.0, 1.0
        Csur = (1 - np.cosh(theta_s * S)) / (np.cosh(theta_s) - 1)
        Cbot = np.sinh(theta_b * (S + 1)) / np.sinh(theta_b) - 1
        mu = (S + 1) ** a * (1 + (a / b) * (1 - (S + 1) ** b))
        return mu * Csur + (1 - mu) * Cbot

    elif Vstretching == 4:
        C = (1 - np.cosh(theta_s * S)) / (np.cosh(theta_s) - 1)
        C = (np.exp(theta_b * C) - 1) / (1 - np.exp(-theta_b))
        return C

    else:
        raise ValueError("Unknown Vstretching")


def sdepth(H, Hc, C, stagger="rho", Vtransform=1):
    """Return depth of rho-points in s-levels

    *H* : arraylike
      Bottom depths [meter, positive]

    *Hc* : scalar
       Critical depth

    *cs_r* : 1D array
       s-level stretching curve

    *stagger* : [ 'rho' | 'w' ]

    *Vtransform* : [ 1 | 2 ]
       defines the transform used, defaults 1 = Song-Haidvogel

    Returns an array with ndim = H.ndim + 1 and
    shape = cs_r.shape + H.shape with the depths of the
    mid-points in the s-levels.

    Typical usage::

    >>> fid = Dataset(roms_file)
    >>> H = fid.variables['h'][:, :]
    >>> C = fid.variables['Cs_r'][:]
    >>> Hc = fid.variables['hc'].getValue()
    >>> z_rho = sdepth(H, Hc, C)

    """
    H = np.asarray(H)
    Hshape = H.shape  # Save the shape of H
    H = H.ravel()  # and make H 1D for easy shape manipulation
    C = np.asarray(C)
    N = len(C)
    outshape = (N,) + Hshape  # Shape of output
    if stagger == "rho":
        S = -1.0 + (0.5 + np.arange(N)) / N  # Unstretched coordinates
    elif stagger == "w":
        S = np.linspace(-1.0, 0.0, N)
    else:
        raise ValueError("stagger must be 'rho' or 'w'")

    if Vtransform == 1:  # Default transform by Song and Haidvogel
        A = Hc * (S - C)[:, None]
        B = np.outer(C, H)
        return (A + B).reshape(outshape)

    elif Vtransform == 2:  # New transform by Shchepetkin
        N = Hc * S[:, None] + np.outer(C, H)
        D = 1.0 + Hc / H
        return (N / D).reshape(outshape)

    else:
        raise ValueError("Unknown Vtransform")


# ------------------------
#   Sampling routines
# ------------------------

# TODO: Unify with s-routine


def vert_level(z_level, X, Y, Z):
    """
    Find vertical level and coefficients for vertical interpolation

    input: X, Y, Z is 3D position, Z positive

    Returns K and A where:

    K is a integer array such that
       0 <= z_level[K] <= Z < z_level[K+1], K = 0, ...

    A is a 2D float array such that
        Z = A*z_level[K] + (1-A)*z_level[K+1]

    """

    # kmax = z_level.shape[0]-1          # Number of vertical

    # K = np.sum(z_w[:, J, I] < -Z, axis=0) - 1
    # K = K.clip(0, kmax-1)
    K = np.searchsorted(z_level, Z) - 1

    A = (z_level[K + 1] - Z) / (z_level[K + 1] - z_level[K])

    A = A.clip(0, 1)

    return K, A


def sample3D(F, X, Y, K, A, method="bilinear"):
    """
    Sample a 3D field on the (sub)grid

    F = 3D field
    S = depth structure matrix
    X, Y = 1D arrays of horizontal grid coordinates
    Z = 1D array of depth [m, positive downwards]

    Everything in rho-points

    F.shape = S.shape = (kmax, jmax, imax)
    S.shape = (kmax, jmax, imax)
    X.shape = Y.shape = Z.shape = (pmax,)

    # Interpolation = 'bilinear' for trilinear Interpolation
    # = 'nearest' for value in 3D grid cell

    """

    # print('sample3D: method =', method)

    # sjekk om 1-A eller A
    if method == "bilinear":
        # Find rho-point as lower left corner
        I = X.astype("int")
        J = Y.astype("int")
        P = X - I
        Q = Y - J
        W000 = (1 - P) * (1 - Q) * A
        W010 = (1 - P) * Q * A
        W100 = P * (1 - Q) * A
        W110 = P * Q * A
        W001 = (1 - P) * (1 - Q) * (1 - A)
        W011 = (1 - P) * Q * (1 - A)
        W101 = P * (1 - Q) * (1 - A)
        W111 = P * Q * (1 - A)

        return (
            W000 * F[K, J, I]
            + W010 * F[K, J + 1, I]
            + W100 * F[K, J, I + 1]
            + W110 * F[K, J + 1, I + 1]
            + W001 * F[K - 1, J, I]
            + W011 * F[K - 1, J + 1, I]
            + W101 * F[K - 1, J, I + 1]
            + W111 * F[K - 1, J + 1, I + 1]
        )

    else:  # method == 'nearest'
        I = X.round().astype("int")
        J = Y.round().astype("int")
        return F[K, J, I]


def sample3DUV(U, V, X, Y, K, A, method="bilinear"):
    return (
        sample3D(U, X + 0.5, Y, K, A, method=method),
        sample3D(V, X, Y + 0.5, K, A, method=method),
    )



"""Output module for the (py)ladim particle tracking model"""

# ------------------------------------
# Bjørn Ådlandsvik <bjorn@imr.no>
# Institute of Marine Research
# 2013-01-04
# ------------------------------------

import os
import logging
import datetime
import re

# from pathlib import Path
import numpy as np
from netCDF4 import Dataset


# Gjør til en iterator
class OutPut:
    def __init__(self, modules, **config):
        self.modules = modules

        logging.info("Initializing output")

        self.filename = config["output_file"]
        self.instance_variables = config["output_instance"]
        self.instance_count = 0
        self.outcount = -1  # No output yet
        self.file_counter = -1  # No file yer
        self.skip_output = config["skip_initial"]
        self.numrec = config["output_numrec"]

        self.outconf = dict(
            output_period=config['output_period'],
            output_format=config['output_format'],
            reference_time=config['reference_time'],
            output_particle=config['output_particle'],
            nc_attributes=config['nc_attributes'].copy(),
            output_instance=config['output_instance'],
        )

        if self.numrec == 0:
            self.multi_file = False
            self.numrec = 999999  # A large number
        else:
            self.multi_file = True
            # If multi_file and output_file ends with number "dddd.nc"
            # start output number with dddd, else use zero
            fname0, ext = os.path.splitext(self.filename)
            # fname = f'{fname0}_{self.file_counter:04d}{ext}'
            patt = r"_[0-9]{4}$"  # '_dddd' at end of string
            m = re.search(patt, fname0)
            # start with number in output_name
            if m:
                self.filename = fname0[:-5] + ext  # Strip the number
                self.file_counter = int(m.group(0)[1:]) - 1

        self.dt = config["dt"]
        self.num_output = config["num_output"]
        self.nc = None  # No open netCD file yet
        # Indicator for lon/lat output
        self.lonlat = (
            "lat" in self.instance_variables or "lon" in self.instance_variables
        )

    # ----------------------------------------------
    def write(self) -> None:
        config = self.outconf
        state = self.modules['state']
        grid = self.modules['grid']

        """Write the model state to NetCDF"""

        # May skip initial output
        if self.skip_output:
            self.skip_output = False
            return

        self.outcount += 1
        t = self.outcount % self.numrec  # in-file record counter

        logging.debug(
            "Writing: timestep, timestamp = {} {}".format(
                state.timestep, state.timestamp
            )
        )

        # Create new file?
        if t == 0:
            # Close old file and open a new
            if self.nc:
                self.nc.close()
            self.file_counter += 1
            self.pstart0 = self.instance_count  # Start of data in the file
            self.nc = self._define_netcdf()
            logging.info(f"Opened output file: {self.nc.filepath()}")

        pcount = len(state)  # Present number of particles
        pstart = self.instance_count

        logging.debug(f"Writing {pcount} particles")

        tdelta = state.timestamp - config["reference_time"]
        seconds = tdelta.astype("m8[s]").astype("int")
        self.nc.variables["time"][t] = float(seconds)

        self.nc.variables["particle_count"][t] = pcount

        # Compute lon, lat if needed
        if self.lonlat:
            lon, lat = grid.xy2ll(state.X, state.Y)

        start = pstart - self.pstart0
        end = pstart + pcount - self.pstart0
        # print("start, end = ", start, end)
        for name in self.instance_variables:
            if name == "lon":
                self.nc.variables["lon"][start:end] = lon
            elif name == "lat":
                self.nc.variables["lat"][start:end] = lat
            else:
                self.nc.variables[name][start:end] = state[name]

        # Update counters
        # self.outcount += 1
        self.instance_count += pcount

        # Flush the data to the file
        self.nc.sync()

        # Close final file
        if self.outcount == self.num_output - 1:
            self.nc.close()

    # -----------------------------------------------
    def _define_netcdf(self) -> Dataset:
        """Define a NetCDF output file"""

        config = self.outconf
        release = self.modules['release']

        # Generate file name
        fname = self.filename
        if self.multi_file:
            # fname0 -> fname0_dddd.nc
            fname0, ext = os.path.splitext(self.filename)
            fname = f"{fname0}_{self.file_counter:04d}{ext}"

        logging.debug(f"Defining output netCDF file: {fname}")
        nc = Dataset(fname, mode="w", format=config["output_format"])
        # --- Dimensions
        nc.createDimension("particle", release.total_particle_count)
        nc.createDimension("particle_instance", None)  # unlimited
        # Sett output-period i config (bruk naturlig enhet)
        # regne om til antall tidsteg og få inn under
        outcount0 = self.outcount
        outcount1 = min(outcount0 + self.numrec, self.num_output)
        nc.createDimension("time", outcount1 - outcount0)

        # ---- Coordinate variable for time
        v = nc.createVariable("time", "f8", ("time",))
        v.long_name = "time"
        v.standard_name = "time"
        timeref = str(config["reference_time"]).replace("T", " ")
        v.units = f"seconds since {timeref}"

        # instance_offset
        v = nc.createVariable("instance_offset", "i", ())
        v.long_name = "particle instance offset for file"

        # Particle count
        v = nc.createVariable("particle_count", "i4", ("time",))
        v.long_name = "number of particles in a given timestep"
        v.ragged_row_count = "particle count at nth timestep"

        # Particle variables
        for name in config["output_particle"]:
            confname = config["nc_attributes"][name]
            if confname["ncformat"][0] == "S":  # text
                length = int(confname["ncformat"][1:])
                lendimname = "len_" + name
                nc.createDimension(lendimname, length)
                v = nc.createVariable(
                    varname=name,
                    datatype="S1",
                    dimensions=("particle", lendimname),
                    zlib=True,
                )
            else:  # Numeric
                v = nc.createVariable(
                    varname=name,
                    datatype=config["nc_attributes"][name]["ncformat"],
                    dimensions=("particle",),
                    zlib=True,
                )
            for attr, value in config["nc_attributes"][name].items():
                if attr != "ncformat":
                    setattr(v, attr, value)

        # Instance variables
        for name in config["output_instance"]:
            v = nc.createVariable(
                varname=name,
                datatype=config["nc_attributes"][name]["ncformat"],
                dimensions=("particle_instance",),
                zlib=True,
            )

            for attr, value in config["nc_attributes"][name].items():
                if attr != "ncformat":
                    setattr(v, attr, value)

        # --- Global attributes
        # Burde ta f.eks. source fra setup
        # hvis andre skulle bruke
        nc.Conventions = "CF-1.5"
        nc.institution = "Institute of Marine Research"
        nc.source = "Lagrangian Advection and Diffusion Model, python version"
        nc.history = "Created by pyladim"
        nc.date = str(datetime.date.today())

        logging.debug("Netcdf output file defined")

        # Save particle variables
        for name in config["output_particle"]:
            var = nc.variables[name]
            if var.datatype == np.dtype("S1"):  # Text
                n = len(nc.dimensions[var.dimensions[-1]])
                A = [
                    list(s[:n].ljust(n))
                    for s in release.particle_variables[name][:]
                ]
                var[:] = np.array(A)
            else:  # Numeric
                nc.variables[name][:] = release.particle_variables[name][:]

        # Set instance offset
        var = nc.variables["instance_offset"]
        var[:] = self.instance_count

        return nc

    def update(self):
        step = self.modules['state'].timestep
        if step % self.outconf['output_period'] == 0:
            self.write()



from collections import namedtuple
import datetime
from typing import Any, List, Dict, Union, Optional
import numpy as np  # type: ignore
import xarray as xr  # type: ignore

from .variable import InstanceVariable, ParticleVariable, arraystr


Timetype = Union[str, np.datetime64, datetime.datetime]
Array = Union[np.ndarray, xr.DataArray]


# --------------------------------------------


Position = namedtuple("Position", "X Y")


class Trajectory:
    """Single particle trajectory"""

    def __init__(self, X: xr.DataArray, Y: xr.DataArray):
        self._data = X, Y

    # For unpacking: X, Y = pf.trajectory(pid=4)
    def __getitem__(self, n):
        return self._data[n]

    @property
    def X(self) -> xr.DataArray:
        return self._data[0]

    @property
    def Y(self) -> xr.DataArray:
        return self._data[1]

    @property
    def time(self) -> np.datetime64:
        return self.X.time

    def __len__(self) -> int:
        return len(self.X.time)


# ---------------------------------------------


class Time:
    """Callable version of time DataArray

    For backwards compability, obsolete
    """

    def __init__(self, ptime):
        self.da = ptime

    def __call__(self, n: int) -> np.datetime64:
        """Prettier version of self[n]"""
        return self.da[n].values.astype("M8[s]")

    def __getitem__(self, arg):
        return self.da[arg]

    def __repr__(self) -> str:
        return arraystr(self.da)

    def __len__(self) -> int:
        return len(self.da)


# --------------------------------------


class ParticleFile:
    def __init__(self, filename: str) -> None:
        ds = xr.open_dataset(filename)
        self.ds = ds
        # End and start of segment with particles at a given time
        self.count = ds.particle_count.values
        self.end = self.count.cumsum()
        self.start = self.end - self.count
        self.num_times = len(self.count)
        self.time = Time(ds.time)
        self.num_particles = int(ds.pid.max()) + 1  # Number of particles

        # Extract instance and particle variables from the netCDF file
        self.instance_variables: List["InstanceVariable"] = []
        self.particle_variables: List["ParticleVariable"] = []
        self.variables: Dict[str, Union["InstanceVariable", "ParticleVariable"]] = {}
        for var in list(self.ds.variables):
            if "particle_instance" in self.ds[var].dims:
                self.instance_variables.append(var)
                self.variables[var] = InstanceVariable(
                    self.ds[var], self.ds.pid, self.ds.time, self.count
                )
            elif "particle" in self.ds[var].dims:
                self.particle_variables.append(var)
                self.variables[var] = ParticleVariable(self.ds[var])

    # For convenience
    def position(self, time: int) -> Position:
        return Position(self.X[time], self.Y[time])

    # For backwards compability
    # Could define ParticleDataset (from file)
    # This could slice and take trajectories og that
    # Could improve speed by computing X and Y at same time
    def trajectory(self, pid: int) -> Trajectory:
        # mypy barks on these two
        # X = self["X"].sel(pid=pid)
        # Y = self["Y"].sel(pid=pid)
        X = InstanceVariable(self.ds["X"], self.ds.pid, self.ds.time, self.count).sel(
            pid=pid
        )
        Y = InstanceVariable(self.ds["Y"], self.ds.pid, self.ds.time, self.count).sel(
            pid=pid
        )
        return Trajectory(X, Y)

    # Obsolete
    def particle_count(self, time: int) -> int:
        return self.count[time]

    def __len__(self) -> int:
        return len(self.time)

    def __getattr__(self, var: str) -> Union[InstanceVariable, ParticleVariable]:
        return self.variables[var]

    def __getitem__(self, var: str) -> Union[InstanceVariable, ParticleVariable]:
        return self.variables[var]

    # Missing: Add global attributes
    def __repr__(self):
        s = "<postladim.ParticleFile>\n"
        s += f"num_times: {self.num_times}, num_particles: {self.num_particles}\n"
        s += f"time: {arraystr(self.time.da)}\n"
        s += f"count: {arraystr(self.count)}\n"
        s += "Instance variables:\n"
        for var in self.instance_variables:
            s += f"  {var:16s} {arraystr(self[var].da)}\n"
        s += "Particle variables:\n"
        for var in self.particle_variables:
            s += f"  {var:16s} {arraystr(self[var].da)}\n"
        s += "Attributes:\n"
        for a, v in self.ds.attrs.items():
            s += f"  {a:16s} {v}\n"
        return s

    def close(self) -> None:
        self.ds.close()

    # Make ParticleFile a context manager
    def __enter__(self):
        return self

    def __exit__(self, atype, value, traceback):
        self.close()


# ----------------------
# Utility functions
# ---------------------



"""
Surface ligth module
"""

# After A. Skartveit & J.A. Olseth,1988
# Varighetstabeller for timevis belysning mot 5 flater
# på 16 norske stasjoner
# Meteorological report series, 1988-7
# University of Bergen

import numpy as np


def light(time, lon, lat, depth=0, extinction_coef=0.2):
    light_0 = surface_light(time, lon, lat)
    return light_0 * np.exp(-extinction_coef * depth)


def surface_light(dtime, lon, lat):
    """Surface light in absence of clouds"""

    RAD = np.pi / 180.0
    DEG = 180 / np.pi
    sin = np.sin
    cos = np.cos
    lat = np.array(lat)
    lon = np.array(lon)
    dtime = np.datetime64(dtime)

    maxlight = 1500  # value between 200 and 2000
    twilight = 5.76

    # --- Time handling ----
    # Convert from datetime64 to datetime
    dtime = dtime.astype(object)
    time_tuple = dtime.timetuple()
    # day of year, original does not consider leap years
    yday = time_tuple.tm_yday
    # hours in UTC (as output from oceanographic model)
    hours = time_tuple.tm_hour

    phi = lat * RAD

    # Compute declineation = delta
    a0 = 0.3979
    a1 = 0.9856 * RAD  # day-1
    a2 = 1.9171 * RAD
    a3 = 0.98112
    sindelta = a0 * sin(a1 * (yday - 80) + a2 * (sin(a1 * yday) - a3))
    cosdelta = (1 - sindelta ** 2) ** 0.5

    # True Sun Time [degrees](=0 with sun in North, 15 deg/hour
    # b0 = 0.4083
    # b1 = 1.7958
    # b2 = 2.4875
    # b3 = 1.0712 * rad   # day-1
    # TST = (hours*15 + lon - b0*np.cos(a1*(yday-80)) -
    #        b1*np.cos(a1*(yday-80)) + b2*np.sin(b3*(yday-80)))

    # TST = 15 * hours  # Recover values from the fortran code

    # Simplified formula
    # correct at spring equinox (yday=80) neglecting +/- 3 deg = 12 min
    TST = hours * 15 + lon

    # Sun height  [degrees]
    # sinheight = sindelta*sin(phi) - cosdelta*cos(phi)*cos(15*hours*rad)
    sinheight = sindelta * sin(phi) - cosdelta * cos(phi) * cos(TST * RAD)
    height = np.arcsin(sinheight) * DEG

    # sine of sun height at noon, h12
    sinh12 = sindelta * sin(phi) + cosdelta * cos(phi)

    # Do we need the surface light?
    # a treshold on sun heigth might be enough

    # Surface light
    slight = np.zeros_like(lat, dtype=float)
    I0 = height >= 0
    I1 = (height >= -6) & (height < 0)
    I2 = (height >= -12) & (height < -6)
    I3 = (height >= -18) & (height < -12)
    I4 = height < -18

    # Is this from Skagseth & Olset ?  Where?
    slight[I0] = maxlight * (sinheight[I0] / sinh12[I0]) + twilight
    slight[I1] = ((twilight - 0.048) / 6) * (6 + height[I1]) + 0.048
    slight[I2] = ((0.048 - 1.15e-4) / 6) * (12 + height[I2]) + 1.15e-4
    slight[I3] = ((1.15e-4 - 1.15e-5) / 6) * (18 + height[I3]) + 1.15e-5
    slight[I4] = 1.15e-5

    return slight



# -*- coding: utf-8 -*-

"""Utilities functions for sampling output from the ROMS

Horizontal sampling
-------------------

:func:`sample2D`
  Sample a 2D field given at rho-points
:func:`sample2DU`
  Sample a 2D field given at u-points
:func:`sample2DV`
  Sample a 2D field given at v-points

"""

# -----------------------------------
# Bjørn Ådlandsvik, bjorn@imr.no
# Institute of Marine Research
# Bergen, Norway
# 2010-09-30
# -----------------------------------

from typing import Tuple
import numpy as np


# ---------------------
def sample2D2(F: np.ndarray, X, Y) -> np.ndarray:
    """Bilinear sample of a 2D field

    *F* : 2D array

    *X*, *Y* : position in grid coordinates, scalars or compatible arrays

    Note reversed axes, for integers i and j we have
      ``sample2D(F, i, j) = F[j,i]``

    Using linear interpolation

    """

    Z = np.add(X, Y)  # Test for compatibility
    if np.isscalar(Z):  # Both X and Y are scalars
        I = int(X)
        J = int(Y)
        P = X - I
        Q = Y - J
    else:
        # Make arrays of common shape
        X0 = X + np.zeros_like(Z)
        Y0 = Y + np.zeros_like(Z)
        I = X0.astype("int")
        J = Y0.astype("int")
        P = X0 - I
        Q = Y0 - J

    W00 = (1 - P) * (1 - Q)
    W01 = (1 - P) * Q
    W10 = P * (1 - Q)
    W11 = P * Q

    return W00 * F[J, I] + W01 * F[J + 1, I] + W10 * F[J, I + 1] + W11 * F[J + 1, I + 1]


# --------------------------------------------------


def sample2DUV(U, V, X, Y) -> Tuple[np.ndarray, np.ndarray]:
    """2D interpolation of velocity"""
    return sample2D(U, X + 0.5, Y), sample2D(V, X, Y + 0.5)


# -------------------------------------------------


def sample2D_masked(F, M, X, Y):
    """Bilinear sample of a 2D field

    F = 2D array, M = mask (=1 on sea, = 0 on land)
    X, Y = position in grid coordinates, scalars or compatible arrays

    Note reversed axes, for integers i and j we have
    sample2D(F, i, j) = F[j,i]

    Using linear interpolation

    """

    masked = True

    Z = np.add(X, Y)  # Test for compatibility
    if np.isscalar(Z):  # Both X and Y are scalars
        I = int(X)
        J = int(Y)
        P = X - I
        Q = Y - J
    else:
        # Make arrays of common shape
        X0 = X + np.zeros_like(Z)
        Y0 = Y + np.zeros_like(Z)
        I = X0.astype("int")
        J = Y0.astype("int")
        P = X0 - I
        Q = Y0 - J

    W00 = M[J, I] * (1 - P) * (1 - Q)
    W01 = M[J + 1, I] * (1 - P) * Q
    W10 = M[J, I + 1] * P * (1 - Q)
    W11 = M[J + 1, I + 1] * P * Q
    if masked:
        W00 = M[J, I] * W00
        W01 = M[J + 1, I] * W01
        W10 = M[J, I + 1] * W10
        W11 = M[J + 1, I + 1] * W11
    SW = W00 + W01 + W10 + W11
    F_interp = np.where(
        SW > 0,
        (W00 * F[J, I] + W01 * F[J + 1, I] + W10 * F[J, I + 1] + W11 * F[J + 1, I + 1])
        / SW,
        0.0,
    )
    # F_interp = np.ma.masked_where(SW == 0, F_interp)

    return F_interp


# ------------------


def sample2D(F, X, Y, mask=None, undef_value=0.0, outside_value=None):
    """Bilinear sample of a 2D field

    *F* : 2D array

    *X*, *Y* : position in grid coordinates, scalars or compatible arrays

    *mask* : if present must be a 2D matrix with 1 at valid
             and zero at non-valid points

    *undef_value* : value to put at undefined points

    *outside_value* : value to return outside the grid
                defaults to None,
                raising ValueError if any points are outside

    Note reversed axes, for integers i and j we have
      ``sample2D(F, i, j) = F[j,i]``

    If jmax, imax = F.shape
    then inside values requires 0 <= x < imax-1, 0 <= y < jmax-1

    Using bilinear interpolation

    """

    # --- Argument checking ---

    # X and Y should be broadcastable to the same shape
    Z = np.add(X, Y)
    # scalar is True if both X and Y are scalars
    scalar = np.isscalar(Z)

    if np.ndim(F) != 2:
        raise ValueError("F must be 2D")
    if mask is not None:
        if mask.shape != F.shape:
            raise ValueError("Must have mask.shape == F.shape")

    jmax, imax = F.shape

    # Broadcast X and Y
    X0 = X + np.zeros_like(Z)
    Y0 = Y + np.zeros_like(Z)
    # Find integer I, J such that
    # 0 <= I <= X < I+1 <= imax-1, 0 <= J <= Y < J+1 <= jmax-1
    # and local increments P and Q
    I = X0.astype("int")
    J = Y0.astype("int")
    P = X0 - I
    Q = Y0 - J

    outside = (X0 < 0) | (X0 >= imax - 1) | (Y0 < 0) | (Y0 >= jmax - 1)
    if np.any(outside):
        if outside_value is None:
            raise ValueError("point outside grid")
        I = np.where(outside, 0, I)
        J = np.where(outside, 0, J)
        # try:
        #    J[outside] = 0
        #    I[outside] = 0
        # except TypeError:    # Zero-dimensional
        #    I = np.array(0)
        #    J = np.array(0)

    # Weights for bilinear interpolation
    W00 = (1 - P) * (1 - Q)
    W01 = (1 - P) * Q
    W10 = P * (1 - Q)
    W11 = P * Q
    SW = 1.0  # Sum of weights

    if mask is not None:
        W00 = mask[J, I] * W00
        W01 = mask[J + 1, I] * W01
        W10 = mask[J, I + 1] * W10
        W11 = mask[J + 1, I + 1] * W11
        SW = W00 + W01 + W10 + W11

    SW = np.where(SW == 0, -1.0, SW)  # Avoid division by zero below
    S = np.where(
        SW <= 0,
        undef_value,
        (W00 * F[J, I] + W01 * F[J + 1, I] + W10 * F[J, I + 1] + W11 * F[J + 1, I + 1])
        / SW,
    )

    # Set in outside_values
    if outside_value:
        S = np.where(outside, outside_value, S)

    #   Scalar input gives scalar output
    if scalar:
        S = float(S)

    return S


# -----------------------------------------


def bilin_inv(f, g, F, G, maxiter=7, tol=1.0e-7):
    """Inverse bilinear interpolation

    f, g : scalars or arrays of same shape
    F, G : 2D arrays of the same shape

    returns x, y : shaped like f and g
    such that F and G linearly interpolated to x, y
    returns f and g

    """

    imax, jmax = F.shape
    if G.shape != (imax, jmax):
        raise ValueError("Shape mismatch in 2D arrays")

    scalar = np.isscalar(f)

    if scalar:
        if not np.isscalar(g):
            raise ValueError("Target values must both be scalars or both arrays")
        # initial guess = mid point
        x = 0.5 * imax
        y = 0.5 * jmax

    else:  # vector target
        f = np.asarray(f)
        g = np.asarray(g)
        fshape = f.shape
        if g.shape != fshape:
            raise ValueError("Target arrays must have the same shape")
        # Make 1D
        # f = f.ravel()
        # g = g.ravel()

        # initial guess
        x = np.zeros_like(f) + 0.5 * imax
        y = np.zeros_like(f) + 0.5 * jmax

    for t in range(maxiter):

        if scalar:
            i = int(x)
            j = int(y)
        else:
            i = x.astype("i")
            j = y.astype("i")

        p, q = x - i, y - j

        # Bilinear estimate of F[x,y] and G[x,y]
        Fs = (
            (1 - p) * (1 - q) * F[i, j]
            + p * (1 - q) * F[i + 1, j]
            + (1 - p) * q * F[i, j + 1]
            + p * q * F[i + 1, j + 1]
        )
        Gs = (
            (1 - p) * (1 - q) * G[i, j]
            + p * (1 - q) * G[i + 1, j]
            + (1 - p) * q * G[i, j + 1]
            + p * q * G[i + 1, j + 1]
        )

        H = (Fs - f) ** 2 + (Gs - g) ** 2
        # print t, H
        if np.all(H < tol):
            break

        # Estimate Jacobi matrix
        Fx = (1 - q) * (F[i + 1, j] - F[i, j]) + q * (F[i + 1, j + 1] - F[i, j + 1])
        Fy = (1 - p) * (F[i, j + 1] - F[i, j]) + p * (F[i + 1, j + 1] - F[i + 1, j])
        Gx = (1 - q) * (G[i + 1, j] - G[i, j]) + q * (G[i + 1, j + 1] - G[i, j + 1])
        Gy = (1 - p) * (G[i, j + 1] - G[i, j]) + p * (G[i + 1, j + 1] - G[i + 1, j])

        # Newton-Raphson step
        # Jinv = np.linalg.inv([[Fx, Fy], [Gx, Gy]])
        # incr = - np.dot(Jinv, [Fs-f, Gs-g])
        # x = x + incr[0], y = y + incr[1]
        det = Fx * Gy - Fy * Gx
        x -= (Gy * (Fs - f) - Fy * (Gs - g)) / det
        y -= (-Gx * (Fs - f) + Fx * (Gs - g)) / det

    return x, y


# ----------------------------



class Legacy_IBM:
    def __init__(self, modules, **config):
        self.modules = modules

        legacy_conf = dict(
            ibm=config,
            dt=config['dt'],
            start_time=config['start_time'],
            output_instance=config['output_instance'],
            nc_attributes=config['nc_attributes'],
        )

        if config["legacy_module"] is not None:
            # Import the module
            import logging
            import sys
            import os
            import importlib
            logging.info("Initializing the IBM")
            sys.path.insert(0, os.getcwd())
            ibm_module = importlib.import_module(config["legacy_module"])
            # Initiate the IBM object
            self.ibm = ibm_module.IBM(legacy_conf)

        else:
            self.ibm = None

    def update(self):
        if self.ibm is not None:
            self.ibm.update_ibm(
                self.modules['grid'],
                self.modules['state'],
                self.modules['forcing'],
            )



"""
General utilities for LADiM
"""

from typing import Any, Dict, List
import numpy as np


def timestep2stamp(config: Dict[str, Any], n: int) -> np.datetime64:
    """Convert from time step number to timestamp"""
    timestamp = config["start_time"] + n * np.timedelta64(config["dt"], "s")
    return timestamp


def timestamp2step(config: Dict[str, Any], timestamp: np.datetime64) -> int:
    """Convert from timestamp to time step number"""
    # mtime = np.datetime64(timestamp)
    dtime = np.timedelta64(timestamp - config["start_time"], "s").astype(int)
    step = dtime // config["dt"]
    return step


# Utility function to test for position in grid
def ingrid(x: float, y: float, subgrid: List[int]) -> bool:
    """Check if position (x, y) is in a subgrid"""
    i0, i1, j0, j1 = subgrid
    return (i0 <= x) & (x <= i1 - 1) & (j0 <= y) & (y <= j1 - 1)



import numpy as np


class TimeStepper:
    def __init__(self, modules, start, stop, step, order=None, seed=None):
        self.order = order or ('forcing', 'release', 'output', 'ibm', 'tracker')
        self.modules = modules
        self.start = np.datetime64(start)
        self.stop = np.datetime64(stop)
        self.step = np.timedelta64(step, 's')
        self.time = None

        if seed is not None:
            np.random.seed(seed)

    def run(self):
        modules = [self.modules[k] for k in self.order if k in self.modules]

        self.time = self.start
        while self.time <= self.stop:
            for m in modules:
                m.update()
            self.time += self.step



import subprocess
import ladim.main
from pathlib import Path
import os
import xarray as xr
import json
import io
import numpy as np


class Test_ladim_script:
    def test_can_show_help_message(self):
        cmd = ['ladim', '--help']
        output = subprocess.run(cmd, capture_output=True)
        assert output.stderr.decode('latin1') == ""
        assert output.stdout.decode('latin1').startswith("usage: ladim")

    def test_can_advect_particles_when_old_version(self):
        curdir = Path.cwd()
        outfile = Path('out.nc')
        testpath = Path(__file__).parent / 'sample_data'

        with open(testpath / 'ladim.yaml') as f:
            conf_str = f.read()

        with open(testpath / 'output.nc_txt', 'r', encoding='utf-8') as f:
            expected = json.load(f)

        dset_dict = None
        try:
            os.chdir(testpath)
            ladim.main(io.StringIO(conf_str))
            dset = xr.load_dataset(str(outfile))
            dset_dict = json.loads(json.dumps(
                obj=dset.to_dict(),
                default=str,
            ))
            for v in ['X', 'Y']:
                d = dset_dict['data_vars'][v]['data']
                dset_dict['data_vars'][v]['data'] = np.round(d, 3).tolist()

        finally:
            try:
                outfile.unlink()
            except IOError:
                pass
            os.chdir(curdir)

        del expected['attrs']['date']
        del dset_dict['attrs']['date']
        assert dset_dict == expected



from ladim import timestepper
import numpy as np


class Test_Solver_run:
    def test_executes_modules_in_specified_order_until_stop(self):
        class MockParticles:
            def __init__(self, m):
                self.modules = m
                self.variables = dict(x=[])

            def __getitem__(self, item):
                return self.variables[item]

            def __repr__(self):
                return str(self.variables)

        class MockSource:
            def __init__(self, m):
                self.modules = m

            def update(self):
                self.modules['state']['x'].append(0)

        class MockOutput:
            def __init__(self, m):
                self.modules = m
                self.record = dict(pid=[], x=[])

            def update(self):
                x = self.modules['state']['x']
                self.record['x'].append(x.copy())
                self.record['pid'].append(list(range(len(x))))

        class MockTransport:
            def __init__(self, m):
                self.modules = m

            def update(self):
                x = self.modules['state']['x']
                x[:] = np.array(x) + 1

        modules = dict()
        modules['output'] = MockOutput(modules)
        modules['tracker'] = MockTransport(modules)
        modules['release'] = MockSource(modules)
        modules['state'] = MockParticles(modules)
        modules['solver'] = timestepper.TimeStepper(
            modules,
            start="2020-01-01",
            stop="2020-01-03",
            step=60*60*24,
            order=['release', 'output', 'tracker'],
        )

        modules['solver'].run()

        assert modules['output'].record['pid'] == [[0], [0, 1], [0, 1, 2]]
        assert modules['output'].record['x'] == [[0], [1, 0], [2, 1, 0]]



import importlib

import ladim.gridforce
import ladim.ibms
import ladim.output
import ladim.release
import ladim.state
import ladim.tracker
import ladim.timestepper

DEFAULT_MODULES = dict(
    grid='ladim.gridforce.Grid',
    forcing='ladim.gridforce.Forcing',
    release='ladim.release.Releaser',
    state='ladim.state.State',
    output='ladim.output.Output',
    ibm='ladim.ibms.IBM',
    tracker='ladim.tracker.Tracker',
    timestepper='ladim.timestepper.TimeStepper',
)


class Model:
    def __init__(self, config):
        module_names = (
            'grid', 'forcing', 'release', 'state', 'output', 'ibm', 'tracker',
            'timestepper',
        )

        self.modules = dict()
        for name in module_names:
            self.add_module(name, config[name])

    def add_module(self, name, conf):
        module_name = conf.get('module', DEFAULT_MODULES[name])
        Module = load_module(module_name)
        self.modules[name] = Module(self, **conf)

    @property
    def grid(self) -> ladim.gridforce.Grid:
        return self.modules.get('grid', None)

    @property
    def forcing(self) -> ladim.gridforce.Forcing:
        return self.modules.get('forcing', None)

    @property
    def release(self) -> ladim.release.Releaser:
        return self.modules.get('release', None)

    @property
    def state(self) -> ladim.state.State:
        return self.modules.get('state', None)

    @property
    def output(self) -> ladim.output.Output:
        return self.modules.get('output', None)

    @property
    def ibm(self) -> ladim.ibms.IBM:
        return self.modules.get('ibm', None)

    @property
    def tracker(self) -> ladim.tracker.Tracker:
        return self.modules.get('tracker', None)

    @property
    def timestepper(self) -> ladim.timestepper.TimeStepper:
        return self.modules.get('timestepper', None)

    def __getitem__(self, item):
        return self.modules[item]

    def __contains__(self, item):
        return item in self.modules

    def run(self):
        self.timestepper.run()

    def close(self):
        for m in self.modules.values():
            if hasattr(m, 'close') and callable(m.close):
                m.close()


def load_module(name):
    pkg, cls = name.rsplit(sep='.', maxsplit=1)
    return getattr(importlib.import_module(pkg), cls)



"""
Configuration module for ladim
"""

# ----------------------------------
# Bjørn Ådlandsvik <bjorn@imr.no>
# Institute of Marine Research
# 2017-01-17
# ----------------------------------

# import datetime
import logging
from typing import Dict, Any
import numpy as np
import yaml
import yaml.parser
from netCDF4 import Dataset, num2date

Config = Dict[str, Any]  # type of the config dictionary


def configure(conf):
    config = read_configuration(conf)
    config = to_modularized_conf(config)
    return config

def configure_ibm(conf: Dict[str, Any]) -> Config:
    """Configure the IBM module

    Input: Raw conf dictionary from configuration file

    Return: Dictionary with IBM configuration

    If an IBM is used, check that module name is present
    Special treatment for the variables item
    Other items are stored for the IBM module
    """

    logging.info("Configuration: IBM")
    if conf is None:  # No ibm section
        return {}
    D = conf.get("ibm")  # Empty ibm section
    if D is None:
        return {}

    # Mandatory: module name (or obsolete ibm_module)
    if "module" not in D:
        if "ibm_module" in D:
            D["module"] = D.pop("ibm_module")
        else:
            logging.error("No IBM module specified")
            raise SystemExit(1)
    logging.info(f'    {"module":15s}: {D["module"]}')

    # The variables item
    if "variables" not in D:
        if "ibm_variables" in D:
            D["variables"] = D.pop("ibm_variables")
        # ibm_variables may live under state (obsolete)
        elif "state" in conf and conf["state"] is not None:
            if "ibm_variables" in conf.get("state", dict()):
                D["variables"] = conf["state"]["ibm_variables"]
        else:
            D["variables"] = []

    for key in D:
        if key != "module":
            logging.info(f"    {key:15s}: {D[key]}")

    return D


def configure_gridforce(conf: Dict[str, Any]) -> Config:
    """Parse gridforce related info and pass on

    Input: raw conf dictionary from configuration file

    Return: dictionary with gridforce configuration

    """
    logging.info("Configuration: gridforce")
    if conf is None:
        logging.error("No gridforce section in configuration file")
        raise SystemExit(1)
    D = conf.get("gridforce")
    if D is None:
        logging.error("Empty gridforce section in configuration file")
        raise SystemExit(1)

    # module is the only mandatory field
    if "module" not in D:
        logging.error("No gridforce module specified")
        raise SystemExit(1)
    logging.info(f'    {"module":15s}: {D["module"]}')

    # Backwards compability (for ROMS.py)
    if "files" in conf and conf["files"] is not None:
        if "grid_file" in conf["files"]:
            # Give grid_file under gridforce highest priority
            if "grid_file" not in D:
                D["grid_file"] = conf["files"]["grid_file"]
        if "input_file" in conf["files"]:
            # Give input_file under gridforce highest priority
            if "input_file" not in D:
                D["input_file"] = conf["files"]["input_file"]

    for key in D:
        if key != "module":
            logging.info(f"    {key:15s}: {D[key]}")

    return D


# ---------------------------------------


def read_configuration(conf) -> Config:
    """The main configuration handling function

    Input: Name of configuration file in yaml format

    Returns: Configuration dictionary

    """

    config: Config = dict()


    # ----------------
    # Time control
    # ----------------
    logging.info("Configuration: Time Control")
    for name in ["start_time", "stop_time"]:
        config[name] = np.datetime64(conf["time_control"][name]).astype("M8[s]")
        logging.info(f"    {name.replace('_', ' '):15s}: {config[name]}")
    # reference_time, default = start_time
    config["reference_time"] = np.datetime64(
        conf["time_control"].get("reference_time", config["start_time"])
    ).astype("M8[s]")
    logging.info(f'    {"reference time":15s}: {config["reference_time"]}')

    # -------------
    # Files
    # -------------
    logging.info("Configuration: Files")
    #logging.info(f'    {"config_stream":15s}: {config_stream}')
    for name in ["particle_release_file", "output_file"]:
        config[name] = conf["files"][name]
        logging.info(f"    {name:15s}: {config[name]}")

    try:
        config["warm_start_file"] = conf["files"]["warm_start_file"]
        config["start"] = "warm"
        logging.info(f'    {"Warm start from":15s}: {config["warm_start_file"]}')
    except KeyError:
        config["start"] = "cold"
        config["warm_start_file"] = ""

    # Override start time for warm start
    if config["start"] == "warm":
        try:
            nc = Dataset(config["warm_start_file"])
        except (FileNotFoundError, OSError):
            logging.error(f"Could not open warm start file,{config['warm_start_file']}")
            raise SystemExit(1)
        tvar = nc.variables["time"]
        # Use last record in restart file
        warm_start_time = np.datetime64(num2date(tvar[-1], tvar.units))
        warm_start_time = warm_start_time.astype("M8[s]")
        config["start_time"] = warm_start_time
        logging.info(f"    Warm start at {warm_start_time}")

        # Variables needed by restart, mightwarm_ be changed
        # default should be instance variables among release variables
        try:
            warm_start_variables = conf["state"]["warm_start_variables"]
        except KeyError:
            warm_start_variables = ["X", "Y", "Z"]
        config["warm_start_variables"] = warm_start_variables

    # --- Time stepping ---
    logging.info("Configuration: Time Stepping")
    # Read time step and convert to seconds
    dt = np.timedelta64(*tuple(conf["numerics"]["dt"]))
    config["dt"] = int(dt.astype("m8[s]").astype("int"))
    config["simulation_time"] = np.timedelta64(
        config["stop_time"] - config["start_time"], "s"
    ).astype("int")
    config["numsteps"] = config["simulation_time"] // config["dt"]
    logging.info(f'    {"dt":15s}: {config["dt"]} seconds')
    logging.info(
        f'    {"simulation time":15s}: {config["simulation_time"] // 3600} hours'
    )
    logging.info(f'    {"number of time steps":15s}: {config["numsteps"]}')

    #  --- Grid ---
    config["gridforce"] = configure_gridforce(conf)

    # --- Forcing ---
    try:
        config["ibm_forcing"] = conf["gridforce"]["ibm_forcing"]
    except (KeyError, TypeError):
        config["ibm_forcing"] = []
    # ibm_forcing used to be a dictionary
    if isinstance(config["ibm_forcing"], dict):
        config["ibm_forcing"] = list(config["ibm_forcing"].keys())
    logging.info(f'    {"ibm_forcing":15s}: {config["ibm_forcing"]}')

    # --- IBM ---

    config["ibm"] = configure_ibm(conf)
    # Make obsolete
    config["ibm_variables"] = config["ibm"].get("variables", [])
    config["ibm_module"] = config["ibm"].get("module")

    # --- Particle release ---
    logging.info("Configuration: Particle Releaser")
    prelease = conf["particle_release"]
    try:
        config["release_type"] = prelease["release_type"]
    except KeyError:
        config["release_type"] = "discrete"
    logging.info(f'    {"release_type":15s}: {config["release_type"]}')
    if config["release_type"] == "continuous":
        config["release_frequency"] = np.timedelta64(
            *tuple(prelease["release_frequency"])
        )
        logging.info(
            f'    {"release_frequency":11s}: {str(config["release_frequency"])}'
        )
    config["release_format"] = conf["particle_release"]["variables"]
    config["release_dtype"] = dict()
    # Map from str to converter
    type_mapping = dict(int=int, float=float, time=np.datetime64, str=str)
    for name in config["release_format"]:
        config["release_dtype"][name] = type_mapping[
            conf["particle_release"].get(name, "float")
        ]
        logging.info(f'    {name:15s}: {config["release_dtype"][name]}')
    config["particle_variables"] = prelease["particle_variables"]

    # --- Model state ---
    # logging.info("Configuration: Model State Variables")

    # -----------------
    # Output control
    # -----------------
    logging.info("Configuration: Output Control")
    try:
        output_format = conf["output_variables"]["format"]
    except KeyError:
        output_format = "NETCDF3_64BIT_OFFSET"
    config["output_format"] = output_format
    logging.info(f'    {"output_format":15s}: {config["output_format"]}')

    # Skip output of initial state, useful for restart
    # with cold start the default is False
    # with warm start, the default is true
    try:
        skip_initial = conf["output_variables"]["skip_initial_output"]
    except KeyError:
        skip_initial = config["start"] == "warm"
    config["skip_initial"] = skip_initial
    logging.info(f"    {'Skip inital output':15s}: {skip_initial}")

    try:
        numrec = conf["output_variables"]["numrec"]
    except KeyError:
        numrec = 0
    config["output_numrec"] = numrec
    logging.info(f'    {"output_numrec":15s}: {config["output_numrec"]}')

    outper = np.timedelta64(*tuple(conf["output_variables"]["outper"]))
    outper = outper.astype("m8[s]").astype("int") // config["dt"]
    config["output_period"] = outper
    logging.info(f'    {"output_period":15s}: {config["output_period"]} timesteps')
    config["num_output"] = 1 + config["numsteps"] // config["output_period"]
    logging.info(f'    {"numsteps":15s}: {config["numsteps"]}')
    config["output_particle"] = conf["output_variables"]["particle"]
    config["output_instance"] = conf["output_variables"]["instance"]
    config["nc_attributes"] = dict()
    for name in config["output_particle"] + config["output_instance"]:
        value = conf["output_variables"][name]
        if "units" in value:
            if value["units"] == "seconds since reference_time":
                timeref = str(config["reference_time"]).replace("T", " ")
                value["units"] = f"seconds since {timeref}"
        config["nc_attributes"][name] = conf["output_variables"][name]
    logging.info("    particle variables")
    for name in config["output_particle"]:
        logging.info(8 * " " + name)
        for item in config["nc_attributes"][name].items():
            logging.info(12 * " " + "{:11s}: {:s}".format(*item))
    logging.info("    particle instance variables")
    for name in config["output_instance"]:
        logging.info(8 * " " + name)
        for item in config["nc_attributes"][name].items():
            logging.info(12 * " " + "{:11s}: {:s}".format(*item))

    # --- Numerics ---

    # dt belongs here, but is already read
    logging.info("Configuration: Numerics")
    try:
        config["advection"] = conf["numerics"]["advection"]
    except KeyError:
        config["advection"] = "RK4"
    logging.info(f'    {"advection":15s}: {config["advection"]}')
    try:
        diffusion = conf["numerics"]["diffusion"]
    except KeyError:
        diffusion = 0.0
    if diffusion > 0:
        config["diffusion"] = True
        config["diffusion_coefficient"] = diffusion
        logging.info(
            f'    {"diffusion coefficient":15s}: {config["diffusion_coefficient"]}'
        )
    else:
        config["diffusion"] = False
        logging.info("    no diffusion")
    config["seed"] = conf['numerics'].get('seed', None)

    return config


def to_modularized_conf(c):
    mconf = dict(
        release=dict(
            module='ladim.release.legacy.ParticleReleaser',
            release_type=c['release_type'],
            release_format=c['release_format'],
            release_dtype=c['release_dtype'],
            start_time=c['start_time'],
            stop_time=c['stop_time'],
            particle_release_file=c['particle_release_file'],
            start=c['start'],
            dt=c['dt'],
            warm_start_file=c['warm_start_file'],
            particle_variables=c['particle_variables'],
            reference_time=c['reference_time'],
            release_frequency=c.get("release_frequency", None),
        ),
        state=dict(
            module='ladim.state.legacy.State',
            particle_variables=c['particle_variables'],
            start_time=c['start_time'],
            dt=c['dt'],
            ibm=dict(variables=c['ibm'].get('variables', [])),
            release_dtype=c['release_dtype'],
            warm_start_file=c['warm_start_file'],
            ibm_forcing=c['ibm_forcing'],
        ),
        grid={
            **c['gridforce'],
            **dict(
                start_time=c['start_time'],
                legacy_module=c['gridforce']['module'],
                module='ladim.gridforce.legacy.Grid',
            ),
        },
        forcing={
            **c['gridforce'],
            **dict(
                start_time=c['start_time'],
                stop_time=c['stop_time'],
                dt=c['dt'],
                legacy_module=c['gridforce']['module'],
                module='ladim.gridforce.legacy.Forcing',
            ),
        },
        output=dict(
            module='ladim.output.legacy.OutPut',
            output_format=c['output_format'],
            skip_initial=c['skip_initial'],
            output_numrec=c['output_numrec'],
            output_period=c['output_period'],
            num_output=c['num_output'],
            output_particle=c['output_particle'],
            output_instance=c['output_instance'],
            nc_attributes=c['nc_attributes'],
            reference_time=c['reference_time'],
            output_file=c['output_file'],
            dt=c['dt'],
        ),
        timestepper=dict(
            start=c['start_time'],
            stop=c['stop_time'],
            step=c['dt'],
            order=('release', 'forcing', 'output', 'tracker', 'ibm', 'state'),
            seed=c.get("seed", None),
        ),
        tracker=dict(
            module='ladim.tracker.legacy.Tracker',
            advection=c['advection'],
            diffusion=c['diffusion'],
            dt=c['dt'],
            ibm_variables=c['ibm_variables'],
            diffusion_coefficient=c.get('diffusion_coefficient', 0),
        ),
        ibm={
            **c['ibm'],
            **dict(
                module='ladim.ibms.legacy.Legacy_IBM',
                dt=c['dt'],
                start_time=c['start_time'],
                nc_attributes=c['nc_attributes'],
                output_instance=c['output_instance'],
                legacy_module=c['ibm'].get('module', None),
            ),
        },
    )
    return mconf


def strdict(d, ind=0):
    s = ""
    for k, v in d.items():
        if isinstance(v, dict):
            s += (" " * ind) + f"{k}:\n"
            s += strdict(v, ind + 2)
        else:
            s += (" " * ind) + f"{k}: {v}\n"

    return s



"""
Main gridforce controller
"""

import os
import sys
import importlib

# from ladim.configuration import config


class Grid:
    def __init__(self, modules, **conf):
        legacy_conf = dict(
            gridforce=conf,
            start_time=conf.get('start_time', None),
        )

        # Allow gridforce module in current directory
        sys.path.insert(0, os.getcwd())
        # Import correct gridforce_module
        # gridforce_module = importlib.import_module(config["gridforce_module"])
        gridforce_module = importlib.import_module(conf["legacy_module"])
        self.grid = gridforce_module.Grid(legacy_conf)
        self.xmin = self.grid.xmin
        self.xmax = self.grid.xmax
        self.ymin = self.grid.ymin
        self.ymax = self.grid.ymax

    def sample_metric(self, X, Y):
        """Sample the metric coefficients"""
        return self.grid.sample_metric(X, Y)

    def sample_depth(self, X, Y):
        """Return the depth of grid cells"""
        return self.grid.sample_depth(X, Y)

    def lonlat(self, X, Y, method=None):
        """Return the longitude and latitude from grid coordinates"""
        return self.grid.lonlat(X, Y, method=method)

    def ingrid(self, X, Y):
        """Returns True for points inside the subgrid"""
        return self.grid.ingrid(X, Y)

    def onland(self, X, Y):
        """Returns True for points on land"""
        return self.grid.onland(X, Y)

    # Error if point outside
    def atsea(self, X, Y):
        """Returns True for points at sea"""
        return self.grid.atsea(X, Y)

    def ll2xy(self, lon, lat):
        return self.grid.ll2xy(lon, lat)

    def xy2ll(self, X, Y):
        return self.grid.xy2ll(X, Y)


class Forcing:
    def __init__(self, modules, **conf):
        self.modules = modules

        grid_ref = GridReference(modules)
        legacy_conf = dict(
            gridforce=conf,
            ibm_forcing=conf.get('ibm_forcing', []),
            start_time=conf.get('start_time', None),
            stop_time=conf.get('stop_time', None),
            dt=conf.get('dt', None),
        )

        # Allow gridforce module in current directory
        sys.path.insert(0, os.getcwd())
        # Import correct gridforce_module
        gridforce_module = importlib.import_module(conf["legacy_module"])
        self.forcing = gridforce_module.Forcing(legacy_conf, grid_ref)
        # self.steps = self.forcing.steps
        # self.U = self.forcing.U
        # self.V = self.forcing.V

    def update(self):
        t = self.modules['state'].timestep
        return self.forcing.update(t)

    def velocity(self, X, Y, Z, tstep=0.0):
        return self.forcing.velocity(X, Y, Z, tstep=tstep)

    def field(self, X, Y, Z, name):
        return self.forcing.field(X, Y, Z, name)

    def close(self):
        return self.forcing.close()


class GridReference:
    def __init__(self, modules):
        self.modules = modules

    def __getattr__(self, item):
        return getattr(self.modules['grid'].grid, item)



# Particle release class

# -------------------
# release.py
# part of LADiM
# --------------------

# ----------------------------------
# Bjørn Ådlandsvik <bjorn@imr.no>
# Institute of Marine Research
# Bergen, Norway
# ----------------------------------

import logging
import numpy as np
import pandas as pd
from typing import Iterator, List

from netCDF4 import Dataset

from ladim.utilities import ingrid


# from .gridforce import Grid


def mylen(df: pd.DataFrame) -> int:
    """Number of rows in a DataFrame,

    A workaround for len() which does not
    have the expected behaviour with itemizing,
    """
    return df.shape[0] if df.ndim > 1 else 1
    # if df.ndim == 1:
    #     return 1
    # else:
    #     return df.shape[0]


class ParticleReleaser(Iterator):
    """Particle Release Class"""

    def __init__(self, modules, **config) -> None:
        self.modules = modules

        start_time = pd.to_datetime(config["start_time"])
        stop_time = pd.to_datetime(config["stop_time"])

        logging.info("Initializing the particle releaser")

        # Read the particle release file
        A = pd.read_csv(
            config["particle_release_file"],
            names=config["release_format"],
            converters=config["release_dtype"],
            sep='\\s+',
        )

        # If no mult column, add a column of ones
        if "mult" not in config["release_format"]:
            A["mult"] = 1

        # Use release_time as index
        A.index = A["release_time"]

        # Conversion from longitude, latitude to grid coordinates
        if "X" not in A.columns or "Y" not in A.columns:
            if "lon" not in A.columns or "lat" not in A.columns:
                logging.critical("Particle release mush have position")
                raise SystemExit(3)
            # else
            X, Y = modules['grid'].ll2xy(A["lon"], A["lat"])
            A["lon"] = X
            A["lat"] = Y
            A.rename(columns={"lon": "X", "lat": "Y"}, inplace=True)

        # Remove everything after simulation stop time
        # A = A[A['release_time'] <= stop_time]   # Use < ?
        A = A[A.index <= stop_time]  # Use < ?
        if len(A) == 0:  # All release after simulation time
            logging.critical("All particles released after similation stop")
            raise SystemExit(3)

        # Optionally, remove everything outside a subgrid
        try:
            subgrid: List[int] = config["grid_args"]["subgrid"]
        except KeyError:
            subgrid = []
        if subgrid:
            lenA = len(A)
            A = A[ingrid(A["X"], A["Y"], subgrid)]
            if len(A) < lenA:
                logging.warning("Ignoring particle release outside subgrid")

        # file_times = A['release_time'].unique()

        # TODO: Make a function of continuous -> discrete
        # Fill out if continuous release
        if config["release_type"] == "continuous":

            # Find first effective release time
            # i.e. the last time <= start_time
            #   and remove too early releases
            # Can be moved out of if-block?
            n = np.sum(A.index <= start_time)
            if n == 0:
                logging.warning("No particles released at simulation start")
                n = 1
            # First effective release:
            # release_time0 = A.iloc[n-1].release_time

            # TODO: Check pandas, better way to delete rows?
            times = A["release_time"]
            try:
                release_time0 = times[times <= pd.to_datetime(start_time)][-1]
            except IndexError:
                release_time0 = times[0]
            A = A[A.index >= release_time0]

            # time0 = file_times[0]
            # time1 = max(file_times[-1], stop_time)
            time0 = A.index[0]
            time1 = max(A.index[-1], pd.Timestamp(stop_time))
            # time1 = max(A['release_time'][-1], stop_time)
            times = np.arange(time0, time1, config["release_frequency"])
            # A = A.reindex(times, method='pad')
            # A['release_time'] = A.index
            # Reindex does not work with non-unique index
            I = A.index.unique()
            J = pd.Series(I, index=I).reindex(times, method="pad")
            M = {i: mylen(A.loc[i]) for i in I}
            A = A.loc[J]
            # Correct time index
            S: List[int] = []
            for t in times:
                S.extend(M[J[t]] * [t])
            A["release_time"] = S
            A.index = S

            # Remove any new instances before start time
            # n = np.sum(A['release_time'] <= start_time)
            # if n == 0:
            #     n = 1
            # A = A.iloc[n-1:]

            # If first release is early set it to start time
            # A['release_time'] = np.maximum(A['release_time'], # tart_time)

        # If discrete, there is an error to have only early releases
        # OK if warm start
        else:  # Discrete
            if A.index[-1] < start_time and config["start"] == "cold":
                logging.error("All particles released before similation start")
                raise SystemExit

        # We are now discrete,
        # remove everything before start time
        A = A[A.index >= start_time]

        # If warm start, no new release at start time (already accounted for)
        if config["start"] == "warm":
            A = A[A.index > start_time]

        # Compute which timestep the release should happen
        timediff = A["release_time"] - config['start_time']
        dt = np.timedelta64(config["dt"], 's')
        rel_tstep = np.int32(timediff / dt)

        # Release times, rounded to nearest time step
        self.times = np.unique(config['start_time'] + rel_tstep * dt)

        logging.info("Number of release times = {}".format(len(self.times)))

        # Compute the release time steps
        rel_time = self.times - config["start_time"]
        rel_time = rel_time.astype("m8[s]").astype("int")
        self.steps = rel_time // config["dt"]

        # Make dataframes for each timeframe
        # self._B = [x[1] for x in A.groupby('release_time')]
        self._B = [x[1] for x in A.groupby(rel_tstep)]

        # Read the particle variables
        self._index = 0  # Index of next release
        self._particle_count = 0  # Particle counter

        # Handle the particle variables initially
        # TODO: Need a test to check that this iw working properly
        pvars = dict()
        for name in config["particle_variables"]:
            dtype = config["release_dtype"][name]
            if dtype == np.datetime64:
                dtype = np.float64
            pvars[name] = np.array([], dtype=dtype)

        # Get particle data from  warm start
        if config["start"] == "warm":
            with Dataset(config["warm_start_file"]) as f:
                # warm_particle_count = len(f.dimensions['particle'])
                warm_particle_count = np.max(f.variables["pid"][:]) + 1
                for name in config["particle_variables"]:
                    pvars[name] = f.variables[name][:warm_particle_count]
        else:
            warm_particle_count = 0

        # initital number of particles
        if config["start"] == "warm":
            init_released = warm_particle_count
        else:
            init_released = 0
        particles_released = [init_released] + [df['mult'].sum() for df in self._B]

        # Loop through the releases, collect particle variable data
        mult = A['mult'].values
        for name in config['particle_variables']:
            val = np.repeat(A[name].values, mult)
            if config['release_dtype'][name] == np.datetime64:
                val = (val - config["reference_time"]) / np.timedelta64(1, 's')
            pvars[name] = np.concatenate((pvars[name], val))

        self.total_particle_count = warm_particle_count + np.sum(mult)
        self.particle_variables = pvars
        logging.info("Total particle count = {}".format(self.total_particle_count))
        self.particles_released = particles_released

        # Export total particle count
        # Ugly to write back to config
        # Easier way to give this piece of information to the
        #     output initialization?
        config["total_particle_count"] = self.total_particle_count

        # Reset the counter after the particle counting
        self._index = 0  # Index of next release
        self._particle_count = warm_particle_count

    def update(self):
        step = self.modules['state'].timestep
        grid = self.modules['grid']
        state = self.modules['state']

        # Extension, allow inactive particles (not moved next time)
        if "active" in state.ibm_variables:
            pass
            # self.active = self.ibm_variables['active']
        else:  # Default = active
            state.active = np.ones_like(state.pid)

        # Surface/bottom boundary conditions
        #     Reflective  at surface
        I = state.Z < 0
        state.Z[I] = -state.Z[I]
        #     Keep just above bottom
        H = grid.sample_depth(state.X, state.Y)
        I = state.Z > H
        state.Z[I] = 0.99 * H[I]

        # Compactify by removing dead particles
        # Could have a switch to avoid this if no deaths
        state.pid = state.pid[state.alive]
        for key in state.instance_variables:
            state[key] = state[key][state.alive]

        if step in self.steps:
            V = next(self)
            self.modules['state'].append(V, self.modules['forcing'])

        # From physics all particles are alive
        # self.alive = np.ones(len(self), dtype="bool")
        state.alive = grid.ingrid(state.X, state.Y)

    def __next__(self) -> pd.DataFrame:
        """Perform the next particle release

           Return a DataFrame with the release info,
           repeating mult times

        """

        # This should not happen
        if self._index >= len(self.times):
            raise StopIteration

        # Skip first release if warm start (should be present in start file)
        # Not always, make better test
        # Moving test to state.py
        # if self._index == 0 and self._particle_count > 0:  # Warm start
        #    return

        # rel_time = self.times[self._index]
        # file_time = self._file_times[self._file_index]

        V = self._B[self._index]
        nnew = V.mult.sum()
        # Workaround, missing repeat method for pandas DataFrame
        V0 = V.to_records(index=False)
        V0 = V0.repeat(V.mult)
        V = pd.DataFrame(V0)
        # Do not need the mult column any more
        V.drop("mult", axis=1, inplace=True)
        # Buffer the new values
        # self.V = V
        # self._file_index += 1

        # Add the new pids
        nnew = len(V)

        pids = pd.Series(
            range(self._particle_count, self._particle_count + nnew), name="pid"
        )
        V = V.join(pids)

        # Update the counters
        self._index += 1
        self._particle_count += len(V)

        return V



#! /usr/bin/env python

"""Main program for running LADiM

Lagrangian Advection and Diffusion Model

"""

# ---------------------------------
# Bjørn Ådlandsvik <bjorn@imr.no>
# Institute of Marine Research
# ---------------------------------

import logging

import ladim

from .configuration import configure
from .model import Model


def main(config_stream, loglevel=logging.INFO):
    """Main function for LADiM"""

    # Logging
    logging.getLogger().setLevel(loglevel)

    # Read configuration
    config = configure(config_stream)

    model = Model(config)
    model.run()
    model.close()


def run():
    import sys
    import argparse
    import logging
    import datetime
    from pathlib import Path

    # ===========
    # Logging
    # ===========

    logging.basicConfig(
        level=logging.INFO,
        format='%(levelname)s:%(module)s - %(message)s')

    # ====================
    # Parse command line
    # ====================

    parser = argparse.ArgumentParser(
        description='LADiM — Lagrangian Advection and Diffusion Model')
    parser.add_argument(
        '-d', '--debug',
        help="Show more information",
        action="store_const", dest="loglevel", const=logging.DEBUG,
        default=logging.INFO)
    parser.add_argument(
        '-s', '--silent',
        help='Show less information',
        action="store_const", dest="loglevel", const=logging.WARNING)
    parser.add_argument('config_file', nargs='?', default='ladim.yaml')

    args = parser.parse_args()

    logging.info(" ================================================")
    logging.info(" === Lagrangian Advection and Diffusion Model ===")
    logging.info(" ================================================\n")

    logging.info(f"ladim path: {ladim.__file__.strip('__init.py__')}")
    logging.info(f"ladim version:  {ladim.__version__}\n")
    logging.info(f"python version:  {sys.version.split()[0]}\n")

    logging.info(f"  Configuration file: {args.config_file}")
    logging.info(f"  loglevel = {logging.getLevelName(args.loglevel)}")

    # =============
    # Sanity check
    # =============

    if not Path(args.config_file).exists():
        logging.critical(f'Configuration file {args.config_file} not found')
        raise SystemExit(1)

    # ===================
    # Run the simulation
    # ===================

    # Start message
    now = datetime.datetime.now().replace(microsecond=0)
    logging.info(f'LADiM simulation starting, wall time={now}')

    fp = open(args.config_file, encoding='utf8')
    ladim.main(config_stream=fp, loglevel=args.loglevel)

    # Reset logging and print final message
    logging.getLogger().setLevel(logging.INFO)
    now = datetime.datetime.now().replace(microsecond=0)
    logging.info(f'LADiM simulation finished, wall time={now}')



# ------------------------------------
# tracker.py
# Part of the LADiM Model
#
# Bjørn Ådlandsvik, <bjorn@imr.no>
# Institute of Marine Research
#
# Licenced under the MIT license
# ------------------------------------

import logging
from typing import Any, Tuple
import numpy as np

from ladim.gridforce.legacy import Grid, Forcing

# from .state import State   # Circular import

Velocity = Tuple[np.ndarray, np.ndarray]
State = Any  # Could not find any better


class Tracker:
    """The physical particle tracking kernel"""

    def __init__(self, modules, **config) -> None:
        self.modules = modules

        logging.info("Initiating the particle tracking")
        self.dt = config["dt"]
        if config["advection"]:
            self.advect = getattr(self, config["advection"])
        else:
            self.advect = None
        # Read from config:
        self.diffusion = config["diffusion"]
        if self.diffusion:
            self.D = config["diffusion_coefficient"]  # [m2.s-1]
        self.active_check = 'active' in config['ibm_variables']

    def update(self):
        self.move_particles(
            self.modules['grid'],
            self.modules['forcing'],
            self.modules['state'],
        )

    def move_particles(self, grid: Grid, forcing: Forcing, state: State) -> None:
        """Move the particles"""

        X, Y = state.X, state.Y
        dx, dy = grid.sample_metric(X, Y)
        self.dx, self.dy = dx, dy
        dt = self.dt
        self.num_particles = len(X)
        # Make more elegant, need not do every time
        # Works for C-grid
        self.xmin = grid.xmin + 0.01
        self.xmax = grid.xmax - 0.01
        self.ymin = grid.ymin + 0.01
        self.ymax = grid.ymax - 0.01

        U = np.zeros(self.num_particles, dtype=float)
        V = np.zeros(self.num_particles, dtype=float)

        # --- Advection ---
        if self.advect:
            Uadv, Vadv = self.advect(forcing, state)
            U += Uadv
            V += Vadv

        # --- Diffusion ---
        if self.diffusion:
            Udiff, Vdiff = self.diffuse()
            U += Udiff
            V += Vdiff

        # --- Move the particles

        # New position, if OK
        X1 = X + U * dt / dx
        Y1 = Y + V * dt / dy

        # Do not move out of grid
        I = ~grid.ingrid(X1, Y1)
        X1[I] = X[I]
        Y1[I] = Y[I]
        # Kill particles trying to move out of the grid
        state.alive[I] = False

        if self.active_check:
            # Do not move inactive particles
            I = state.active < 1
            X1[I] = X[I]
            Y1[I] = Y[I]

        # Land, boundary treatment. Do not move the particles
        # Consider a sequence of different actions
        # I = (grid.ingrid(X1, Y1)) & (grid.atsea(X1, Y1))
        I = grid.atsea(X1, Y1)
        # I = True
        X[I] = X1[I]
        Y[I] = Y1[I]

        state.X = X
        state.Y = Y

    def EF(self, forcing: Forcing, state: State) -> Velocity:
        """Euler-Forward advection"""

        X, Y, Z = state["X"], state["Y"], state["Z"]
        # dt = self.dt
        # pm, pn = grid.sample_metric(X, Y)

        U, V = forcing.velocity(X, Y, Z)

        return U, V

    def RK2a(self, forcing: Forcing, state: State) -> Velocity:
        """Runge-Kutta second order = Heun scheme"""

        X, Y, Z = state["X"], state["Y"], state["Z"]
        dt = self.dt

        U, V = forcing.velocity(X, Y, Z)
        X1 = X + 0.5 * U * dt / self.dx
        Y1 = Y + 0.5 * V * dt / self.dy

        U, V = forcing.velocity(X1, Y1, Z, tstep=0.5)
        return U, V

    def RK2b(self, forcing: Forcing, state: State) -> Velocity:
        """Runge-Kutta second order = Heun scheme

        This version does not sample velocities outside the grid
        """

        X, Y, Z = state["X"], state["Y"], state["Z"]
        dt = self.dt

        U, V = forcing.velocity(X, Y, Z)
        X1 = X + 0.5 * U * dt / self.dx
        Y1 = Y + 0.5 * V * dt / self.dy
        X1.clip(self.xmin, self.xmax, out=X1)
        Y1.clip(self.ymin, self.ymax, out=Y1)

        U, V = forcing.velocity(X1, Y1, Z, tstep=0.5)
        return U, V

    RK2 = RK2b

    def RK4a(self, forcing: Forcing, state: State) -> Velocity:
        """Runge-Kutta fourth order advection"""

        X, Y, Z = state["X"], state["Y"], state["Z"]
        dt = self.dt
        dx, dy = self.dx, self.dy

        U1, V1 = forcing.velocity(X, Y, Z, tstep=0.0)
        X1 = X + 0.5 * U1 * dt / dx
        Y1 = Y + 0.5 * V1 * dt / dy

        U2, V2 = forcing.velocity(X1, Y1, Z, tstep=0.5)
        X2 = X + 0.5 * U2 * dt / dx
        Y2 = Y + 0.5 * V2 * dt / dy

        U3, V3 = forcing.velocity(X2, Y2, Z, tstep=0.5)
        X3 = X + U3 * dt / dx
        Y3 = Y + V3 * dt / dy

        U4, V4 = forcing.velocity(X3, Y3, Z, tstep=1.0)

        U = (U1 + 2 * U2 + 2 * U3 + U4) / 6.0
        V = (V1 + 2 * V2 + 2 * V3 + V4) / 6.0

        return U, V

    def RK4b(self, forcing: Forcing, state: State) -> Velocity:
        """Runge-Kutta fourth order advection

        This version does not sample velocities outside the grid

        """

        X, Y, Z = state["X"], state["Y"], state["Z"]
        dt = self.dt
        dx, dy = self.dx, self.dy
        xmin, xmax, ymin, ymax = self.xmin, self.xmax, self.ymin, self.ymax

        U1, V1 = forcing.velocity(X, Y, Z, tstep=0.0)
        X1 = X + 0.5 * U1 * dt / dx
        Y1 = Y + 0.5 * V1 * dt / dy
        X1.clip(xmin, xmax, out=X1)
        Y1.clip(ymin, ymax, out=Y1)

        U2, V2 = forcing.velocity(X1, Y1, Z, tstep=0.5)
        X2 = X + 0.5 * U2 * dt / dx
        Y2 = Y + 0.5 * V2 * dt / dy
        X2.clip(xmin, xmax, out=X2)
        Y2.clip(ymin, ymax, out=Y2)

        U3, V3 = forcing.velocity(X2, Y2, Z, tstep=0.5)
        X3 = X + U3 * dt / dx
        Y3 = Y + V3 * dt / dy
        X3.clip(xmin, xmax, out=X3)
        Y3.clip(ymin, ymax, out=Y3)

        U4, V4 = forcing.velocity(X3, Y3, Z, tstep=1.0)

        U = (U1 + 2 * U2 + 2 * U3 + U4) / 6.0
        V = (V1 + 2 * V2 + 2 * V3 + V4) / 6.0

        return U, V

    RK4 = RK4b

    def diffuse(self) -> Velocity:
        """Random walk diffusion"""

        # Diffusive velocity
        stddev = (2 * self.D / self.dt) ** 0.5
        U = stddev * np.random.normal(size=self.num_particles)
        V = stddev * np.random.normal(size=self.num_particles)

        return U, V



import numpy as np
import matplotlib.pyplot as plt
from netCDF4 import Dataset
import roppy
from roppy.mpl_util import landmask
from particlefile import ParticleFile

# ---------------
# User settings
# ---------------

# Files
# particle_file = '../output/pyladim_out.nc'
particle_file = "../output/streak.nc"
roms_file = "../input/ocean_avg_0014.nc"

# Subgrid definition
i0, j0 = 70, 80
i1, j1 = 150, 133

t = 96  # time step 31

# Length scale (in grid units)
sigma = 3.14

# ----------------

# ROMS grid, plot domain

f0 = Dataset(roms_file)
g = roppy.SGrid(f0, subgrid=(i0, i1, j0, j1))

#
# print particle_file
pf = ParticleFile(particle_file)

# Ntimes = pf.nFrames

X, Y = pf.get_position(t)
tstring = pf.get_time(t)


# # Compute weigths based on land
# N = len(X)
# for k in range(N):
#     A = np.zeros((j1-j0, i1-i0))
#     for j in range(j0, j1):
#         for i in range(i0, i1):
#             A[j-j0, i-i0] += g.mask_rho * np.exp

A = np.zeros((j1 - j0, i1 - i0))
L2 = 2 * sigma
for j in range(j0, j1):
    for i in range(i0, i1):
        A[j - j0, i - i0] = np.sum(np.exp(-((i - X) / L2) ** 2 - ((j - Y) / L2) ** 2))
A /= np.pi * L2 ** 2

print(np.sum(A))

# Create a figure

fig = plt.figure(figsize=(12, 8))
ax = fig.add_subplot(1, 1, 1)

# Make background map
# cmap = plt.get_cmap('Blues')
# h = ax.contourf(g.X, g.Y, g.h, cmap=cmap, alpha=0.3)
# fig.colorbar(h)

h = ax.contourf(g.X, g.Y, A)
fig.colorbar(h)

landmask(g, (0.6, 0.8, 0.0))

# Plot iparticle distribution
h = ax.plot(X, Y, ".", color="black", markeredgewidth=0, lw=0.5)
ax.set_title(tstring)

# Show the results
plt.axis("image")
plt.show()



import numpy as np


class Grid:
    def __init__(self, config):
        self.xmin = -np.inf
        self.xmax = np.inf
        self.ymin = -np.inf
        self.ymax = np.inf

    def sample_depth(self, X, Y):
        return np.zeros(np.shape(X), dtype='i2')

    def sample_metric(self, X, Y):
        m = np.ones(np.shape(X), dtype='i2')
        return m, m

    def lonlat(self, X, Y, method="bilinear"):
        return X, Y

    def ingrid(self, X, Y):
        return np.ones(np.shape(X), dtype=bool)

    def onland(self, X, Y):
        return np.zeros(np.shape(X), dtype=bool)

    def atsea(self, X, Y):
        return np.ones(np.shape(X), dtype=bool)

    def xy2ll(self, X, Y):
        return X, Y

    def ll2xy(self, lon, lat):
        return lon, lat


class Forcing:
    def __init__(self, config, grid):
        pass

    def update(self, t):
        pass

    def close(self):
        pass

    def velocity(self, X, Y, Z, tstep=0, method="bilinear"):
        u = np.ones(np.shape(X))
        return u, u

    def field(self, X, Y, Z, name):
        return np.zeros(np.shape(X))



"""
Class for the state of the model
"""

import sys
import os
import importlib
import logging
from typing import Any, Dict, Sized  # mypy

import numpy as np
from netCDF4 import Dataset, num2date

from ladim.tracker.legacy import Tracker
from ladim.gridforce.legacy import Grid, Forcing

# ------------------------

Config = Dict[str, Any]


class State(Sized):
    """The model variables at a given time"""

    def __init__(self, modules, **config):
        self.modules = modules

        logging.info("Initializing the model state")

        self.timestep = 0
        self.timestamp = config["start_time"].astype("datetime64[s]")
        self.dt = np.timedelta64(config["dt"], "s")
        self.position_variables = ["X", "Y", "Z"]
        if "ibm" in config and "variables" in config["ibm"]:
            self.ibm_variables = config["ibm"]["variables"]
        else:
            self.ibm_variables = config.get("ibm_variables", [])
        self.ibm_forcing = config["ibm_forcing"]
        self.particle_variables = config["particle_variables"]
        self.instance_variables = self.position_variables + [
            var for var in self.ibm_variables if var not in self.particle_variables
        ]

        self.pid = np.array([], dtype=int)
        for name in self.instance_variables:
            setattr(self, name, np.array([], dtype=float))

        for name in self.particle_variables:
            setattr(self, name, np.array([], dtype=config["release_dtype"][name]))

        self.dt = config["dt"]
        self.alive = []

        # self.num_particles = len(self.X)
        self.nnew = 0  # Modify with warm start?

        if config["warm_start_file"]:
            self.warm_start(config, self.modules['grid'])

    def __getitem__(self, name: str) -> None:
        return getattr(self, name)

    def __setitem__(self, name: str, value: Any) -> None:
        return setattr(self, name, value)

    def __len__(self) -> int:
        return len(getattr(self, "X"))

    def append(self, new: Dict[str, Any], forcing: Forcing) -> None:
        """Append new particles to the model state"""
        nnew = len(new["pid"])
        self.pid = np.concatenate((self.pid, new["pid"]))
        for name in self.instance_variables:
            if name in new:
                self[name] = np.concatenate((self[name], new[name]))
            elif name in self.ibm_forcing:
                # Take values as Z must be a numpy array
                self[name] = np.concatenate(
                    (self[name], forcing.field(new["X"], new["Y"], new["Z"].values, name))
                )
            else:  # Initialize to zero
                self[name] = np.concatenate((self[name], np.zeros(nnew)))
        self.nnew = nnew

    def update(self):
        """Update the model state to the next timestep"""
        self.timestep += 1
        self.timestamp += np.timedelta64(self.dt, "s")

        if self.timestamp.astype("int") % 3600 == 0:  # New hour
            logging.info("Model time = {}".format(self.timestamp.astype("M8[h]")))

    def warm_start(self, config: Config, grid: Grid) -> None:
        """Perform a warm (re)start"""

        warm_start_file = config["warm_start_file"]
        try:
            f = Dataset(warm_start_file)
        except FileNotFoundError:
            logging.critical(f"Can not open warm start file: {warm_start_file}")
            raise SystemExit(1)

        logging.info("Reading warm start file")
        # Using last record in file
        tvar = f.variables["time"]
        warm_start_time = np.datetime64(num2date(tvar[-1], tvar.units))
        # Not needed anymore, explicitly set in configuration
        # if warm_start_time != config['start_time']:
        #    print("warm start time = ", warm_start_time)
        #    print("start time      = ", config['start_time'])
        #    logging.error("Warm start time and start time differ")
        #    raise SystemExit(1)

        pstart = f.variables["particle_count"][:-1].sum()
        pcount = f.variables["particle_count"][-1]
        self.pid = f.variables["pid"][pstart : pstart + pcount]
        # Give error if variable not in restart file
        for var in config["warm_start_variables"]:
            logging.debug(f"Reading {var} from warm start file")
            self[var] = f.variables[var][pstart : pstart + pcount]

        # Remove particles near edge of grid
        I = grid.ingrid(self["X"], self["Y"])
        self.pid = self.pid[I]
        for var in config["warm_start_variables"]:
            self[var] = self[var][I]



def configure(module_conf):
    import yaml

    # Handle variations of input config type
    if isinstance(module_conf, dict):
        config_dict = module_conf
    else:
        config_dict = yaml.safe_load(module_conf)

    if 'version' in config_dict:
        return _versioned_configure(config_dict)
    else:
        return _legacy_configure(config_dict)


def _versioned_configure(config_dict):
    return config_dict


def _legacy_configure(config_dict):
    from .legacy import configure as legacy_configure
    return legacy_configure(config_dict)



import numpy as np  # type:ignore
import xarray as xr  # type: ignore
from typing import Union, Optional, Tuple, List

Array = Union[List[float], np.ndarray, xr.DataArray]
Limits = Union[Tuple[int, int], Tuple[int, int, int, int]]


def cellcount(
    X: Array, Y: Array, W: Optional[Array] = None, grid_limits: Optional[Limits] = None
) -> xr.DataArray:
    """Count the (weighted) number of particles in grid cells

    Parameters
    ----------
    X, Y : 1D arrays, length = n
        Particle position in grid coordinates
    W : 1D array, length = n
        Weight of particles, default=None for unweighted
    grid_limits : 4-tuple (i0, i1, j0, j1) or 2-tuple (i1, j1)
        Limitation of grid to consider,
        If 2-tuple, i0 and j0 default to zero.
        Default=None gives the bounding box of the particle positions

    Returns
    -------
    C : 2D xarray.DataArray, shape = (j1-j0, i1-i0)
        Particle counts

    Note: particles outside the grid limits are silently ignored

    """

    # Subgrid specification

    i0: int
    i1: int
    j0: int
    j1: int

    if grid_limits is None:
        i0 = int(round(np.min(X)))
        i1 = int(round(np.max(X))) + 1
        j0 = int(round(np.min(Y)))
        j1 = int(round(np.max(Y))) + 1
    elif len(grid_limits) == 2:
        i1, j1 = grid_limits
        i0, j0 = 0, 0
    elif len(grid_limits) == 4:
        i0, i1, j0, j1 = grid_limits
    else:
        raise TypeError("Illegal grid_limits")

    # Count
    x_edges = np.arange(i0 - 0.5, i1)
    y_edges = np.arange(j0 - 0.5, j1)
    if W is None:
        C = np.histogram2d(np.asarray(Y), np.asarray(X), bins=[y_edges, x_edges])
    else:
        C = np.histogram2d(
            np.asarray(Y), np.asarray(X), weights=np.asarray(W), bins=[y_edges, x_edges]
        )

    coords = dict(Y=np.arange(j0, j1), X=np.arange(i0, i1))
    C = xr.DataArray(C[0], coords=list(coords.items()), dims=coords.keys())
    # mypy does not like the line below
    # C = xr.DataArray(C[0], coords=coords, dims=coords.keys())

    return C
